""" Main postprocessing module that starts all other modules. """
# clean
import os
import sys
from typing import Any, Optional

from hisim.postprocessing import reportgenerator
from hisim.postprocessing import charts
from hisim import log
from hisim import utils
from hisim.postprocessingoptions import PostProcessingOptions
from hisim import loadtypes as lt
from hisim.postprocessing.chart_singleday import ChartSingleDay
from hisim.postprocessing.compute_kpis import compute_kpis
from hisim.postprocessing.system_chart import SystemChart
from hisim.component import ComponentOutput
from hisim.postprocessing.postprocessing_datatransfer import PostProcessingDataTransfer


class PostProcessor:

    """ Core Post processor class. """

    @utils.measure_execution_time
    def __init__(self):
        """ Initializes the post processing. """
        self.dirname: str

    def set_dir_results(self, dirname: Optional[str] = None) -> None:
        """ Sets the results directory. """
        if dirname is None:
            raise ValueError("No results directory name was defined.")
        self.dirname = dirname

    @utils.measure_execution_time
    def plot_sankeys(self, ppdt: PostProcessingDataTransfer) -> None:
        """ For plotting the sankeys. """
        for i_display_name in [name for name, display_name in lt.DisplayNames.__members__.items()]:
            my_sankey = charts.SankeyHISIM(name=i_display_name,
                                           units=lt.Units.ANY,
                                           directorypath=ppdt.simulation_parameters.result_directory,
                                           time_correction_factor=ppdt.time_correction_factor)
            my_sankey.plot(data=ppdt.all_outputs)
        if any(component_output.component_name == "HeatPump" for component_output in ppdt.all_outputs):
            my_sankey = charts.SankeyHISIM(name="HeatPump",
                                           units=lt.Units.ANY,
                                           directorypath=ppdt.simulation_parameters.result_directory,
                                           time_correction_factor=ppdt.time_correction_factor)
            my_sankey.plot_heat_pump(data=ppdt.all_outputs)
        if any(component_output.component_name == "Building" for component_output in ppdt.all_outputs):
            my_sankey = charts.SankeyHISIM(name="Building",
                                           units=lt.Units.ANY,
                                           directorypath=ppdt.simulation_parameters.result_directory,
                                           time_correction_factor=ppdt.time_correction_factor)
            my_sankey.plot_building(data=ppdt.all_outputs)

    @utils.measure_execution_time
    @utils.measure_memory_leak
    def run(self, ppdt: PostProcessingDataTransfer) -> None:  # noqa: MC0001
        """ Runs the main post processing. """
        # Define the directory name
        log.information("Main post processing function")

        # Check whether HiSim is running in a docker container
        docker_flag = os.getenv("HISIM_IN_DOCKER_CONTAINER", "false")
        if docker_flag.lower() in ("true", "yes", "y", "1"):
            # Charts etc. are not needed when executing HiSim in a container. Allow only csv files and KPI.
            allowed_options_for_docker = {PostProcessingOptions.EXPORT_TO_CSV, PostProcessingOptions.COMPUTE_KPI}
            # Of all specified options, select those that are allowed
            valid_options = list(set(ppdt.post_processing_options) & allowed_options_for_docker)
            if len(valid_options) < len(ppdt.post_processing_options):
                # At least one invalid option was set
                ppdt.post_processing_options = valid_options
                log.warning("Hisim is running in a docker container. Disabled invalid postprocessing options.")

        report = reportgenerator.ReportGenerator(dirpath=ppdt.simulation_parameters.result_directory)
        days = {"month": 0, "day": 0}
        if PostProcessingOptions.PLOT_LINE in ppdt.post_processing_options:
            log.information("Making line plots.")
            self.make_line_plots(ppdt)
        if PostProcessingOptions.PLOT_CARPET in ppdt.post_processing_options:
            log.information("Making carpet plots.")
            self.make_carpet_plots(ppdt)
        if PostProcessingOptions.PLOT_SINGLE_DAYS in ppdt.post_processing_options:
            log.information("Making single day plots.")
            self.make_single_day_plots(days, ppdt)
        if PostProcessingOptions.PLOT_BAR_CHARTS in ppdt.post_processing_options:
            log.information("Making bar charts.")
            self.make_bar_charts(ppdt)
        # Plot sankey
        if PostProcessingOptions.PLOT_SANKEY in ppdt.post_processing_options:
            log.information("Making sankey plots.")
            self.make_sankey_plots()
        # Export all results to CSV
        if PostProcessingOptions.EXPORT_TO_CSV in ppdt.post_processing_options:
            log.information("Making CSV exports.")
            self.make_csv_export(ppdt)
        if PostProcessingOptions.GENERATE_PDF_REPORT in ppdt.post_processing_options:
            log.information("Making PDF report.")
            self.write_components_to_report(ppdt, report)
        # Export all results to
        if PostProcessingOptions.COMPUTE_KPI in ppdt.post_processing_options:
            log.information("Computing KPIs")
            self.compute_kpis(ppdt, report)
        if PostProcessingOptions.MAKE_NETWORK_CHARTS in ppdt.post_processing_options:
            log.information("Computing Network Charts")
            self.make_network_charts(ppdt)

        # only a single day has been calculated. This gets special charts for debugging.
        if PostProcessingOptions.PLOT_SPECIAL_TESTING_SINGLE_DAY in ppdt.post_processing_options and len(ppdt.results) == 1440:
            log.information("Making special single day plots for a single day calculation for testing.")
            self.make_special_one_day_debugging_plots(ppdt)

        # Open file explorer
        if PostProcessingOptions.OPEN_DIRECTORY_IN_EXPLORER in ppdt.post_processing_options:
            log.information("opening the explorer.")
            self.open_dir_in_file_explorer(ppdt)
        log.information("Finished main post processing function")

    def make_network_charts(self, ppdt: PostProcessingDataTransfer) -> None:
        """ Generates the network charts that show the connection of the elements. """
        systemchart = SystemChart(ppdt)
        systemchart.make_chart()

    def make_special_one_day_debugging_plots(self, ppdt: PostProcessingDataTransfer) -> None:
        """ Makes special plots for debugging if only a single day was calculated."""
        for index, output in enumerate(ppdt.all_outputs):
            if output.full_name == "Dummy # Residence Temperature":
                my_days = ChartSingleDay(output=output.full_name,
                                         units=output.unit,
                                         directorypath=ppdt.simulation_parameters.result_directory,
                                         time_correction_factor=ppdt.time_correction_factor,
                                         data=ppdt.results.iloc[:, index],
                                         day=0,
                                         month=0,
                                         output2=ppdt.results.iloc[:, 11])
            else:
                my_days = ChartSingleDay(output=output.full_name,
                                         units=output.unit,
                                         directorypath=ppdt.simulation_parameters.result_directory,
                                         time_correction_factor=ppdt.time_correction_factor,
                                         data=ppdt.results.iloc[:, index],
                                         day=0,
                                         month=0)
            my_days.plot(close=True)

    def make_csv_export(self, ppdt: PostProcessingDataTransfer) -> None:
        """ Exports all data to CSV. """
        log.information("exporting to csv")
        self.export_results_to_csv(ppdt)

    def make_sankey_plots(self) -> None:
        """ Makes Sankey plots. Needs work. """
        log.information("plotting sankeys")
        # TODO:   self.plot_sankeys()

    def make_bar_charts(self, ppdt: PostProcessingDataTransfer) -> None:
        """ Make bar charts. """
        for index, output in enumerate(ppdt.all_outputs):
            my_bar = charts.BarChart(output=output.full_name,

                                     units=output.unit,
                                     dirpath=ppdt.simulation_parameters.result_directory,
                                     time_correction_factor=ppdt.time_correction_factor)
            my_bar.plot(data=ppdt.results_monthly.iloc[:, index])

    def make_single_day_plots(self, days: Any, ppdt: PostProcessingDataTransfer) -> None:
        """ Makes plots for selected days. """
        for index, output in enumerate(ppdt.all_outputs):
            my_days = ChartSingleDay(output=output.full_name,
                                     units=output.unit,
                                     directorypath=ppdt.simulation_parameters.result_directory,
                                     time_correction_factor=ppdt.time_correction_factor,
                                     day=days["day"],
                                     month=days["month"],
                                     data=ppdt.results.iloc[:, index])
            my_days.plot(close=True)

    def make_carpet_plots(self, ppdt: PostProcessingDataTransfer) -> None:
        """ Make carpet plots. """
        for index, output in enumerate(ppdt.all_outputs):
            log.trace("Making carpet plots")
            my_carpet = charts.Carpet(output=output.full_name,

                                      units=output.unit,
                                      directorypath=ppdt.simulation_parameters.result_directory,
                                      time_correction_factor=ppdt.time_correction_factor)
            my_carpet.plot(xdims=int(
                (ppdt.simulation_parameters.end_date - ppdt.simulation_parameters.start_date).days), data=ppdt.results.iloc[:, index])

    @utils.measure_memory_leak
    def make_line_plots(self, ppdt: PostProcessingDataTransfer) -> None:
        """ Makes the line plots."""
        for index, output in enumerate(ppdt.all_outputs):
            my_line = charts.Line(output=output.full_name,
                                  units=output.unit,
                                  directorypath=ppdt.simulation_parameters.result_directory,
                                  time_correction_factor=ppdt.time_correction_factor)
            my_line.plot(data=ppdt.results.iloc[:, index], units=output.unit)
            del my_line

    @utils.measure_execution_time
    def export_results_to_csv(self, ppdt: PostProcessingDataTransfer) -> None:
        """ Exports the results to a CSV file. """
        for column in ppdt.results:
            ppdt.results[column].to_csv(os.path.join(ppdt.simulation_parameters.result_directory,
                                                     f"{column.split(' ', 3)[2]}_{column.split(' ', 3)[0]}.csv"), sep=",", decimal=".")
        for column in ppdt.results_monthly:
            csvfilename = os.path.join(ppdt.simulation_parameters.result_directory,
                                       f"{column.split(' ', 3)[2]}_{column.split(' ', 3)[0]}_monthly.csv")
            header = [f"{column.split('[', 1)[0]} - monthly ["f"{column.split('[', 1)[1]}"]
            ppdt.results_monthly[column].to_csv(csvfilename, sep=",", decimal=".", header=header)

    def write_to_report(self, text: Any, report: reportgenerator.ReportGenerator) -> None:
        """ Writes a single line to the report. """
        report.open()
        report.write(text)
        report.close()

    def compute_kpis(self, ppdt: PostProcessingDataTransfer, report: reportgenerator.ReportGenerator) -> None:
        """ Computes KPI's and writes them to report and csv. """
        kpi_compute_return = compute_kpis(results=ppdt.results, all_outputs=ppdt.all_outputs, simulation_parameters=ppdt.simulation_parameters)
        lines = kpi_compute_return
        self.write_to_report(text=lines, report=report)

    def write_components_to_report(self, ppdt: PostProcessingDataTransfer, report: reportgenerator.ReportGenerator) -> None:
        """ Writes information about the components used in the simulation to the simulation report. """
        report.open()
        for wrapped_component in ppdt.wrapped_components:
            if hasattr(wrapped_component.my_component, "write_to_report"):
                component_content = wrapped_component.my_component.write_to_report()
            else:
                raise ValueError("Component is missing write_to_report_function: " + wrapped_component.my_component.component_name)
            if isinstance(component_content, list) is False:
                component_content = [component_content]
            if isinstance(component_content, str) is True:
                component_content = [component_content]
            report.write(component_content)
        all_output_names = []
        output: ComponentOutput
        for output in ppdt.all_outputs:
            all_output_names.append(output.full_name + " [" + output.unit + "]")
        report.write(["### All Outputs"])
        report.write(all_output_names)
        report.close()

    def open_dir_in_file_explorer(self, ppdt: PostProcessingDataTransfer) -> None:
        """ Opens files in given path.

        The keyword darwin is used for supporting macOS,
        xdg-open will be available on any unix client running X.
        """
        if sys.platform == "win32":
            os.startfile(os.path.realpath(ppdt.simulation_parameters.result_directory))  # noqa: B606
        else:
            log.information("Not on Windows. Can't open explorer.")

    def export_sankeys(self):
        """ Exports Sankeys plots.

        ToDo: implement
        """
        pass  # noqa: unnecessary-pass



""" Contains various utility functions and utility classes. """
# clean
import gc
import os
import inspect
import hashlib
import json

from timeit import default_timer as timer
from typing import Any, Dict, Tuple
from functools import wraps

import psutil

from hisim.simulationparameters import SimulationParameters
from hisim import log

__authors__ = "Noah Pflugradt, Vitor Hugo Bellotto Zago"
__copyright__ = "Copyright 2021-2022, FZJ-IEK-3 "
__license__ = "MIT"
__version__ = "1"
__maintainer__ = "Noah Pflugradt"
__email__ = "n.pflugradt@fz-juelich.de"
__status__ = "development"


def get_input_directory() -> str:
    """ Gets the absolute path to the inputs directory. """
    return os.path.join(hisim_abs_path, "inputs")


# Retrieves hisim directory absolute path
hisim_abs_path = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))  # type: ignore
hisim_inputs = os.path.join(hisim_abs_path, "inputs")
hisim_results = os.path.join(hisim_abs_path, "results")
hisim_postprocessing_img = os.path.join(hisim_abs_path, "postprocessing", "report")  # noqa

HISIMPATH: Dict[str, Any] = {"results": hisim_results,
                             "inputs": hisim_inputs,
                             "cache_dir": os.path.join(hisim_abs_path, "inputs", "cache"),
                             "cache_indices": os.path.join(hisim_abs_path, "inputs", "cache", "cache_indices.json"),
                             "cfg": os.path.join(hisim_abs_path, "inputs", "cfg.json"),
                             "utsp_results": os.path.join(hisim_results, "Results"),
                             "utsp_example_results": os.path.join(hisim_inputs, "LPGResults_for_tests", "Results"),
                             "utsp_reports": os.path.join(hisim_results, "Reports"),
                             "utsp_example_reports": os.path.join(hisim_inputs, "LPGResults_for_tests", "Reports"),
                             "housing": os.path.join(hisim_inputs,
                                                     "housing",
                                                     "data_processed",
                                                     "episcope-tabula.csv"),
                             "occupancy": {"CH01": {"number_of_residents": [os.path.join(hisim_inputs,
                                                                                         "loadprofiles",
                                                                                         "electrical-warmwater-presence-load_1-family",
                                                                                         "data_processed",
                                                                                         "BodilyActivityLevel.High.HH1.json"),
                                                                            os.path.join(hisim_inputs,
                                                                                         "loadprofiles",
                                                                                         "electrical-warmwater-presence-load_1-family",
                                                                                         "data_processed",
                                                                                         "BodilyActivityLevel.Low.HH1.json")],
                                                    "electricity_consumption": os.path.join(hisim_inputs,
                                                                                            "loadprofiles",
                                                                                            "electrical-warmwater-presence-load_1-family",
                                                                                            "data_processed",
                                                                                            "SumProfiles.HH1.Electricity.csv"),
                                                    "water_consumption": os.path.join(hisim_inputs,
                                                                                      "loadprofiles",
                                                                                      "electrical-warmwater-presence-load_1-family",
                                                                                      "data_processed",
                                                                                      "SumProfiles.HH1.Warm Water.csv")},
                                           "CHR01 Couple both at Work": {"number_of_residents": [os.path.join(hisim_inputs,
                                                                                         "loadprofiles",
                                                                                         "electrical-warmwater-presence-load_1-family",
                                                                                         "data_processed",
                                                                                         "BodilyActivityLevel.High.HH1.json"),
                                                                            os.path.join(hisim_inputs,
                                                                                         "loadprofiles",
                                                                                         "electrical-warmwater-presence-load_1-family",
                                                                                         "data_processed",
                                                                                         "BodilyActivityLevel.Low.HH1.json")],
                                                    "electricity_consumption": os.path.join(hisim_inputs,
                                                                                            "loadprofiles",
                                                                                            "electrical-warmwater-presence-load_1-family",
                                                                                            "data_processed",
                                                                                            "SumProfiles.HH1.Electricity.csv"),
                                                    "water_consumption": os.path.join(hisim_inputs,
                                                                                      "loadprofiles",
                                                                                      "electrical-warmwater-presence-load_1-family",
                                                                                      "data_processed",
                                                                                      "SumProfiles.HH1.Warm Water.csv")}},
                             "photovoltaic": {"modules": os.path.join(hisim_inputs,
                                                                      "photovoltaic",
                                                                      "data_processed",
                                                                      "sandia_modules.csv"),
                                              "inverters": os.path.join(hisim_inputs,
                                                                        "photovoltaic",
                                                                        "data_processed",
                                                                        "sandia_inverters.csv")},
                             "chp_system": os.path.join(hisim_inputs,
                                                        "chp_system"),

                             "smart_appliances": os.path.join(hisim_inputs,
                                                              "smart_devices",
                                                              "data_processed",
                                                              "smart_devices.json"),
                             "frank_data": os.path.join(hisim_inputs,
                                                        "loadprofiles",
                                                        "electrical-spaceheating-warmwater-photovoltaic_1-household",
                                                        "data_raw",
                                                        "VDI 4655"),

                             "report": os.path.join(hisim_abs_path, "results", "report.pdf"),
                             "advanced_battery": {
                                 "parameter": os.path.join(hisim_abs_path, "inputs", "advanced_battery", "parameter",
                                                           "PerModPAR.xlsx"),
                                 "reference_case": os.path.join(hisim_abs_path, "inputs", "advanced_battery",
                                                                "reference_case", "ref_case_data.npz"),
                                 "siemens_junelight": os.path.join(hisim_abs_path, "inputs", "advanced_battery",
                                                                   "Siemens_Junelight.npy")},
                             "LoadProfileGenerator_export_directory": os.path.join(os.path.join("D:", os.sep, "Work")),
                             "bat_parameter": os.path.join(hisim_abs_path, "inputs", "advanced_battery",
                                                           "Siemens_Junelight.npy"),
                             "modular_household": os.path.join(hisim_abs_path, "modular_household")}


def load_smart_appliance(name):  # noqa
    """ Loads file for a single smart appliance by name. """
    with open(HISIMPATH["smart_appliances"], encoding="utf-8") as filestream:
        data = json.load(filestream)
    return data[name]


def get_cache_file(component_key: str, parameter_class: Any, my_simulation_parameters: SimulationParameters) -> Tuple[bool, str]:  # noqa
    """ Gets a cache path for a given parameter set.

    This will generate a file path based on any dataclass_json.
    It works by turning the class into a json string, hashing the string and then using that as filename.
    The idea is to have a unique file path for every possible configuration.
    """
    json_str = parameter_class.to_json()
    if my_simulation_parameters is None:
        raise ValueError("Simulation parameters was none.")
    simulation_parameter_str = my_simulation_parameters.get_unique_key()
    json_str = json_str + simulation_parameter_str
    if len(json_str) < 5:
        raise Exception("Empty json detected for caching. This is a bug.")
    json_str_encoded = json_str.encode('utf-8')
    # Johanna Ganglbauer: python told me "TypeError: openssl_sha256() takes at most 1 argument (2 given)",
    # I removed the second input argument "usedforsecurity=False" and it works - maybe I need to update the hashlib package?
    sha_key = hashlib.sha256(json_str_encoded).hexdigest()
    filename = component_key + "_" + sha_key + ".cache"
    cache_dir_path = os.path.join(hisim_abs_path, "inputs", "cache")
    cache_absolute_filepath = os.path.join(hisim_abs_path, "inputs", "cache", filename)
    if not os.path.isdir(cache_dir_path):
        os.mkdir(cache_dir_path)
    if os.path.isfile(cache_absolute_filepath):
        return True, cache_absolute_filepath
    return False, cache_absolute_filepath


def load_export_load_profile_generator(target):  # noqa
    """ Returns the paths for the SQL exported files from the Load Profile Generator. """
    targetpath = os.path.join(HISIMPATH["LoadProfileGenerator_export_directory"], target)
    if os.path.exists(targetpath):
        lpg_export_path = {"electric_vehicle": [os.path.join(targetpath, "Results.HH1.sqlite"), os.path.join(targetpath, "Results.General.sqlite")]}
        return lpg_export_path
    raise ValueError("Target export from Load Profile Generator does not exist")


def measure_execution_time(my_function):  # noqa
    """ Utility function that works as decorator for measuring execution time. """
    @wraps(my_function)
    def function_wrapper_for_measuring_execution_time(*args, **kwargs):
        """ Inner function for the time measuring utility decorator. """
        start = timer()
        result = my_function(*args, **kwargs)
        end = timer()
        diff = end - start
        log.profile("Executing " + my_function.__module__ + "." + my_function.__name__ + " took " + f"{diff:1.2f}" + " seconds")
        return result

    return function_wrapper_for_measuring_execution_time


def measure_memory_leak(my_function):  # noqa
    """ Utility function that works as decorator for measuring execution time. """
    @wraps(my_function)
    def function_wrapper_for_measuring_memory_leak(*args, **kwargs):
        """ Inner function for the time measuring utility decorator. """
        process = psutil.Process(os.getpid())
        rss_by_psutil_start = process.memory_info().rss / (1024 * 1024)
        result = my_function(*args, **kwargs)
        rss_by_psutil_end = process.memory_info().rss / (1024 * 1024)
        gc.collect()
        diff = rss_by_psutil_end - rss_by_psutil_start
        log.trace("Executing " + my_function.__module__ + "." + my_function.__name__ + " leaked " + f"{diff:1.2f}" + " MB")
        return result

    return function_wrapper_for_measuring_memory_leak


def measure_memory_leak_with_error(my_function):  # noqa
    """ Utility function that works as decorator for measuring execution time. """
    @wraps(my_function)
    def function_wrapper_for_measuring_memory_leak(*args, **kwargs):
        """ Inner function for the time measuring utility decorator. """
        process = psutil.Process(os.getpid())
        rss_by_psutil_start = process.memory_info().rss / (1024 * 1024)
        result = my_function(*args, **kwargs)
        rss_by_psutil_end = process.memory_info().rss / (1024 * 1024)
        gc.collect()
        diff = rss_by_psutil_end - rss_by_psutil_start
        log.information(
            "Executing " + my_function.__module__ + "." + my_function.__name__ + " leaked " + f"{diff:1.2f}" + " MB")
        if diff > 100:
            raise ValueError("Lost over 100MB of memory during the function call")
        return result

    return function_wrapper_for_measuring_memory_leak


def deprecated(message):
    """ Decorator for marking a function as deprecated. """
    def deprecated_decorator(func):
        """ Decorator. """
        def deprecated_func(*args, **kwargs):
            """ Core function. """
            log.warning(f"{func.__name__} is a deprecated function. {message}")
            return func(*args, **kwargs)
        return deprecated_func
    return deprecated_decorator



""" Logging functionality for all of HiSim. """
# clean
from enum import IntEnum

LOGGING_LEVEL = 3


class LogPrio(IntEnum):

    """ Define a logging priority. """

    ERROR = 1
    WARNING = 2
    INFORMATION = 3
    DEBUG = 4
    PROFILE = 5
    TRACE = 6


def error(message: str) -> None:
    """ Log an error message. """
    log(LogPrio.ERROR, message)


def warning(message: str) -> None:
    """ Log a warning message. """
    log(LogPrio.WARNING, message)


def information(message: str) -> None:
    """ Log a information message. """
    log(LogPrio.INFORMATION, message)


def trace(message: str) -> None:
    """ Log a trace message. """
    log(LogPrio.TRACE, message)


def debug(message: str) -> None:
    """ Log a debug message. """
    log(LogPrio.DEBUG, message)


def profile(message: str) -> None:
    """ Log a profile message. """
    log(LogPrio.PROFILE, message)
    log_profile_file(message)


def log(prio: int, message: str) -> None:
    """ Write and print a log message. """
    # if(prio < LogPrio.Debug):
    prio_string: str
    if prio == LogPrio.ERROR:
        prio_string = "ERR"
    elif prio == LogPrio.WARNING:
        prio_string = "WRN"
    elif prio == LogPrio.INFORMATION:
        prio_string = "IFO"
    elif prio == LogPrio.DEBUG:
        prio_string = "DBG"
    elif prio == LogPrio.PROFILE:
        prio_string = "PRF"
    elif prio == LogPrio.TRACE:
        prio_string = "TRC"
    else:
        raise ValueError("Unknown log priority: " + str(prio))
    if prio <= LOGGING_LEVEL:
        print(str(prio_string) + ":" + message)
    with open('hisim_simulation.log', 'a', encoding="utf-8") as filestream:
        filestream.write(message + "\n")


def log_profile_file(message: str) -> None:
    """ Write log message to logfile. """
    with open('profiling_timeuse.log', 'a', encoding="utf-8") as filestream:
        filestream.write(message + "\n")



# Generic/Built-in
from typing import Any
# Owned
import copy
import numpy as np

from hisim import dynamic_component
from hisim import component as cp
from hisim import loadtypes as lt
from hisim.simulationparameters import SimulationParameters
from hisim import log
from hisim import utils
from typing import List
from dataclasses import dataclass
from dataclasses_json import dataclass_json
@dataclass_json
@dataclass
class ControllerHeatConfig:
    temperature_storage_target_warm_water: float = 50
    temperature_storage_target_heating_water: float = 35
    temperature_storage_target_hysteresis_ww: float = 45
    temperature_storage_target_hysteresis_hw: float = 35


# class ControllerState
class ControllerState:
    """
    Save State if Heater Components were supposed to run
    in last timestep (Control_Signal).
    Saves timestep of hysteresis of storages and  the
    changing target temperature of storages. The
    target temperature in states changes when target
    temperature of storage is reached.
    """

    def __init__(self, control_signal_gas_heater: float, control_signal_chp: float, control_signal_heat_pump: float,
                 temperature_storage_target_ww_C: float, temperature_storage_target_hw_C: float, timestep_of_hysteresis_ww: int,
                 timestep_of_hysteresis_hw: int) -> None:
        self.control_signal_gas_heater: float = control_signal_gas_heater
        self.control_signal_chp: float = control_signal_chp
        self.control_signal_heat_pump: float = control_signal_heat_pump
        self.temperature_storage_target_ww_C: float = temperature_storage_target_ww_C
        self.temperature_storage_target_hw_C: float = temperature_storage_target_hw_C
        self.timestep_of_hysteresis_ww: int = timestep_of_hysteresis_ww
        self.timestep_of_hysteresis_hw: int = timestep_of_hysteresis_hw

    def clone(self) -> "ControllerState":
        return ControllerState(control_signal_gas_heater=self.control_signal_gas_heater, control_signal_chp=self.control_signal_chp,
                               control_signal_heat_pump=self.control_signal_heat_pump,
                               temperature_storage_target_ww_C=self.temperature_storage_target_ww_C,
                               temperature_storage_target_hw_C=self.temperature_storage_target_hw_C,
                               timestep_of_hysteresis_ww=self.timestep_of_hysteresis_ww, timestep_of_hysteresis_hw=self.timestep_of_hysteresis_hw)


class ControllerHeat(cp.Component):
    """
    Controlls energy flows for heat demand.
    Heat Demand will be controlled by the storage temperature.
    For this storage provides heat for a load profile or for the
    building. As well Heat Demand can be simulated without storage.
    """

    # Inputs
    StorageTemperatureHeatingWater = "StorageTemperatureHeatingWater"
    StorageTemperatureWarmWater = "StorageTemperatureWarmWater"
    ResidenceTemperature = "ResidenceTemperature"
    ThermalDemandBuilding = "ThermalDemandBuilding"

    # Outputs
    ControlSignalGasHeater = "ControlSignalGasHeater"
    ControlSignalChp = "ControlSignalChp"
    ControlSignalHeatPump = "ControlSignalHeatPump"
    ControlSignalChooseStorage = "ControlSignalChooseStorage"

    @utils.measure_execution_time
    def __init__(self,
                 my_simulation_parameters: SimulationParameters,
                 config: ControllerHeatConfig) -> None:
        super().__init__(name="ControllerHeat", my_simulation_parameters=my_simulation_parameters)

        self.temperature_storage_target_warm_water = config.temperature_storage_target_warm_water
        self.temperature_storage_target_heating_water = config.temperature_storage_target_heating_water
        self.temperature_storage_target_hysteresis_hw = config.temperature_storage_target_hysteresis_hw
        self.temperature_storage_target_hysteresis_ww = config.temperature_storage_target_hysteresis_ww

        self.state = ControllerState(control_signal_heat_pump=0,
                                     control_signal_gas_heater=0,
                                     control_signal_chp=0,
                                     temperature_storage_target_ww_C=self.temperature_storage_target_warm_water,
                                     temperature_storage_target_hw_C=self.temperature_storage_target_heating_water,
                                     timestep_of_hysteresis_ww=0,
                                     timestep_of_hysteresis_hw=0)
        self.previous_state = self.state.clone()

        ###Inputs
        self.temperature_storage_warm_water: cp.ComponentInput = self.add_input(self.component_name,
                                                                                self.StorageTemperatureWarmWater,
                                                                                lt.LoadTypes.WATER,
                                                                                lt.Units.CELSIUS,
                                                                                False)
        self.temperature_storage_heating_water: cp.ComponentInput = self.add_input(self.component_name,
                                                                                   self.StorageTemperatureHeatingWater,
                                                                                   lt.LoadTypes.WATER,
                                                                                   lt.Units.CELSIUS,
                                                                                   False)
        self.temperature_residence: cp.ComponentInput = self.add_input(self.component_name,
                                                                       self.ResidenceTemperature,
                                                                       lt.LoadTypes.TEMPERATURE,
                                                                       lt.Units.CELSIUS,
                                                                       False)

        # Outputs
        self.control_signal_gas_heater: cp.ComponentOutput = self.add_output(object_name=self.component_name,
                                                                             field_name=self.ControlSignalGasHeater,
                                                                             load_type=lt.LoadTypes.ANY,
                                                                             unit=lt.Units.PERCENT,
                                                                             sankey_flow_direction=False)
        self.control_signal_chp: cp.ComponentOutput = self.add_output(object_name=self.component_name,
                                                                      field_name=self.ControlSignalChp,
                                                                      load_type=lt.LoadTypes.ANY,
                                                                      unit=lt.Units.PERCENT,
                                                                      sankey_flow_direction=False)
        self.control_signal_heat_pump: cp.ComponentOutput = self.add_output(object_name=self.component_name,
                                                                            field_name=self.ControlSignalHeatPump,
                                                                            load_type=lt.LoadTypes.ANY,
                                                                            unit=lt.Units.PERCENT,
                                                                            sankey_flow_direction=False)
        self.control_signal_choose_storage: cp.ComponentOutput = self.add_output(object_name=self.component_name,
                                                                                 field_name=self.ControlSignalChooseStorage,
                                                                                 load_type=lt.LoadTypes.ANY,
                                                                                 unit=lt.Units.ANY,
                                                                                 sankey_flow_direction=False)

    @staticmethod
    def get_default_config() -> ControllerHeatConfig:
        config = ControllerHeatConfig(temperature_storage_target_warm_water=50,
                                      temperature_storage_target_heating_water=35,
                                      temperature_storage_target_hysteresis_ww=45,
                                      temperature_storage_target_hysteresis_hw=35)
        return config

    def build(self, mode: Any) -> None :
        self.mode = mode

    def write_to_report(self) -> None:
        pass
    def i_prepare_simulation(self) -> None:
        """ Prepares the simulation. """
        pass
    def i_save_state(self) -> None:
        self.previous_state = self.state.clone()

    def i_restore_state(self) -> None:
        self.state = self.previous_state.clone()

    def i_doublecheck(self, timestep: int, stsv: cp.SingleTimeStepValues) -> None:
        pass

    # Simulates and defines the control signals to heat up storages
    # work as a 2-point Ruler with Hysteresis
    def simulate_storage(self, delta_temperature: float,
                         stsv: cp.SingleTimeStepValues,
                         timestep: int, temperature_storage: float,
                         temperature_storage_target: float, temperature_storage_target_hysteresis: float,
                         temperature_storage_target_C: float, timestep_of_hysteresis: int) -> Any:
        control_signal_chp: float = 0
        control_signal_gas_heater: float = 0
        control_signal_heat_pump: float = 0
        temperature_storage_target_C = temperature_storage_target_C
        timestep_of_hysteresis = timestep_of_hysteresis

        max_temperature_limit = 5
        if temperature_storage > 0:
            if delta_temperature > max_temperature_limit:
                control_signal_heat_pump = 1
                control_signal_chp = 1
                control_signal_gas_heater = 1
                temperature_storage_target_C = temperature_storage_target


            elif delta_temperature > 0 and delta_temperature <= max_temperature_limit:
                control_signal_heat_pump = 1
                control_signal_chp = 1
                control_signal_gas_heater = 1

                if self.state.control_signal_chp < control_signal_chp:
                    control_signal_chp = 1
                elif self.state.control_signal_gas_heater < control_signal_gas_heater:
                    control_signal_gas_heater = 1
                temperature_storage_target_C = temperature_storage_target

                # Storage warm enough. Try to turn off Heaters
            elif delta_temperature <= 0:
                if temperature_storage_target_C == temperature_storage_target and timestep_of_hysteresis != timestep:
                    temperature_storage_target_C = temperature_storage_target_hysteresis
                    timestep_of_hysteresis = timestep
                elif temperature_storage_target_C != temperature_storage_target and timestep_of_hysteresis != timestep:
                    control_signal_heat_pump = 0
                    control_signal_gas_heater = 0
                    control_signal_chp = 0

        self.state.control_signal_gas_heater = control_signal_gas_heater
        self.state.control_signal_chp = control_signal_chp
        self.state.control_signal_heat_pump = control_signal_heat_pump
        stsv.set_output_value(self.control_signal_heat_pump, control_signal_heat_pump)
        stsv.set_output_value(self.control_signal_gas_heater, control_signal_gas_heater)
        stsv.set_output_value(self.control_signal_chp, control_signal_chp)

        return temperature_storage_target_C, \
               timestep_of_hysteresis

    def i_simulate(self, timestep: int, stsv: cp.SingleTimeStepValues, force_convergence: bool) -> None:
        if force_convergence:
            return
        #######HEAT########
        # Logic of regulating HeatDemand:
        # First heat up WarmWaterStorage->more important, than heat up HeatingWater
        # But only one Storage can be heated up in a TimeStep!
        # Simulate WarmWater
        delta_temperature_ww = self.state.temperature_storage_target_ww_C - stsv.get_input_value(
            self.temperature_storage_warm_water)
        delta_temperature_hw = self.state.temperature_storage_target_hw_C - stsv.get_input_value(
            self.temperature_storage_heating_water)
        if stsv.get_input_value(self.temperature_storage_warm_water) == 0 and stsv.get_input_value(
                self.temperature_storage_heating_water) != 0:
            control_signal_choose_storage = 2
        elif stsv.get_input_value(self.temperature_storage_warm_water) != 0 and stsv.get_input_value(
                self.temperature_storage_heating_water) == 0:
            control_signal_choose_storage = 1
        else:
            # Choose which Storage should be heated up
            if delta_temperature_ww >= 0 and delta_temperature_hw >= 0:
                if delta_temperature_hw <= delta_temperature_ww:
                    control_signal_choose_storage = 1
                else:
                    control_signal_choose_storage = 2
            elif delta_temperature_ww < 0 and delta_temperature_hw < 0:
                if delta_temperature_hw <= delta_temperature_ww:
                    control_signal_choose_storage = 1
                else:
                    control_signal_choose_storage = 2
            elif delta_temperature_ww <= 0 and delta_temperature_hw >= 0:
                control_signal_choose_storage = 2
            elif delta_temperature_ww >= 0 and delta_temperature_hw <= 0:
                control_signal_choose_storage = 1

        # Heats up storage
        if control_signal_choose_storage == 1:
            self.state.temperature_storage_target_ww_C, self.state.timestep_of_hysteresis_ww = self.simulate_storage(
                stsv=stsv,
                delta_temperature=delta_temperature_ww,
                timestep=timestep,
                temperature_storage=stsv.get_input_value(self.temperature_storage_warm_water),
                temperature_storage_target=self.temperature_storage_target_warm_water,
                temperature_storage_target_hysteresis=self.temperature_storage_target_hysteresis_ww,
                temperature_storage_target_C=self.state.temperature_storage_target_ww_C,
                timestep_of_hysteresis=self.state.timestep_of_hysteresis_ww)
        elif control_signal_choose_storage == 2:
            delta_temperature_hw = self.state.temperature_storage_target_hw_C - stsv.get_input_value(
                self.temperature_storage_heating_water)
            self.state.temperature_storage_target_hw_C, self.state.timestep_of_hysteresis_hw = self.simulate_storage(
                stsv=stsv,
                delta_temperature=delta_temperature_hw,
                timestep=timestep,
                temperature_storage=stsv.get_input_value(self.temperature_storage_heating_water),
                temperature_storage_target=self.temperature_storage_target_heating_water,
                temperature_storage_target_hysteresis=self.temperature_storage_target_hysteresis_hw,
                temperature_storage_target_C=self.state.temperature_storage_target_hw_C,
                timestep_of_hysteresis=self.state.timestep_of_hysteresis_hw)

        stsv.set_output_value(self.control_signal_choose_storage, control_signal_choose_storage)



""" Defines the component class and helpers.

The component class is the base class for all other components.
"""
# clean
from __future__ import annotations
from typing import List, Optional, Dict, Any, Type

import typing
import dataclasses as dc
from dataclasses import dataclass
from dataclass_wizard import JSONWizard

# Package

from hisim.simulationparameters import SimulationParameters
from hisim import loadtypes as lt
from hisim import log
from hisim.sim_repository import SimRepository


@dataclass
class ConfigBase(JSONWizard):

    """ Base class for all configurations. """

    name: str

    def __init__(self):
        """ Initializes. """
        pass

    @classmethod
    def get_main_classname(cls):
        """ Returns the fully qualified class name for the class that is getting configured. Used for Json. """
        raise NotImplementedError("Missing a definition of the ")

    @classmethod
    def get_config_classname(cls):
        """ Gets the class name. Helper function for default connections. """
        return cls.__module__ + "." + cls.__name__

    def get_string_dict(self) -> List[str]:
        """ Turns the config into a str list for the report. """
        my_dict = self.to_dict()
        my_list = []
        if len(my_dict) > 0:
            for entry in my_dict.items():
                my_list.append(entry[0] + ": " + str(entry[1]))
        return my_list


@dataclass
class ComponentConnection:

    """ Used in the component class for defining a connection. """

    target_input_name: str
    source_class_name: str
    source_output_name: str
    source_instance_name: Optional[str] = None


class ComponentOutput:  # noqa: too-few-public-methods

    """ Used in the component class for defining an output. """

    def __init__(self, object_name: str, field_name: str, load_type: lt.LoadTypes, unit: lt.Units,
                 postprocessing_flag: Optional[List[Any]] = None,
                 sankey_flow_direction: Optional[bool] = None):
        """ Defines a component output. """
        self.full_name: str = object_name + " # " + field_name
        self.component_name: str = object_name  # ComponentName
        self.field_name: str = field_name
        self.display_name: str = field_name
        self.load_type: lt.LoadTypes = load_type
        self.unit: lt.Units = unit
        self.global_index: int = -1
        self.postprocessing_flag: Optional[List[Any]] = postprocessing_flag
        self.sankey_flow_direction: Optional[bool] = sankey_flow_direction

    def get_pretty_name(self):
        """ Gets a pretty name for a component output. """
        return self.component_name + " - " + self.display_name + " [" + self.load_type + " - " + self.unit + "]"


class ComponentInput:  # noqa: too-few-public-methods

    """ Used in the component class for defining an input. """

    def __init__(self, object_name: str, field_name: str, load_type: lt.LoadTypes, unit: lt.Units, mandatory: bool):
        """ Initializes a component input. """
        self.fullname: str = object_name + " # " + field_name
        self.component_name: str = object_name
        self.field_name: str = field_name
        self.loadtype: lt.LoadTypes = load_type
        self.unit: lt.Units = unit
        self.global_index: int = -1
        self.src_object_name: Optional[str] = None
        self.src_field_name: Optional[str] = None
        self.source_output: Optional[ComponentOutput] = None
        self.is_mandatory = mandatory


class SingleTimeStepValues:

    """ Contains the values for a single time step. """

    def __init__(self, number_of_values: int):
        """ Initializes a new single time step values class. """
        self.values = [0.0] * number_of_values

    def copy_values_from_other(self, other):
        """ Copy all values from a single time step values. """
        self.values = other.values[:]

    def clone(self):
        """ Makes a copy of the current object. """
        newstsv = SingleTimeStepValues(len(self.values))
        newstsv.values = self.values[:]
        return newstsv

    def get_input_value(self, component_input: ComponentInput) -> float:
        """ Gets a value for an input from the single time step values. """
        if component_input.source_output is None:
            return 0
        return self.values[component_input.source_output.global_index]

    def set_output_value(self, output: ComponentOutput, value: float) -> None:
        """ Sets a single output value in the single time step values array. """
        self.values[output.global_index] = value

    def is_close_enough_to_previous(self, previous_values: "SingleTimeStepValues") -> bool:
        """ Checks if the values are sufficiently similar to another array. """
        count = len(self.values)
        for i in range(count):
            if abs(previous_values.values[i] - self.values[i]) > 0.0001:
                return False
        return True

    def get_differences_for_error_msg(self, previous_values: Any, outputs: List[ComponentOutput]) -> str:
        """ Gets a pretty error message for the differences between two time steps. """
        count = len(self.values)
        error_msg = ""
        for i in range(count):
            if abs(previous_values.values[i] - self.values[i]) > 0.0001:
                error_msg += outputs[i].get_pretty_name() + " previously: " + str(
                    previous_values.values[i]) + " currently: " + str(self.values[i])
        return error_msg


class Component:

    """ Base class for all components. """

    @classmethod
    def get_classname(cls):
        """ Gets the class name. Helper function for default connections. """
        return cls.__name__

    @classmethod
    def get_full_classname(cls):
        """ Gets the class name. Helper function for default connections. """
        return cls.__module__ + "." + cls.__name__

    def __init__(self, name: str, my_simulation_parameters: SimulationParameters) -> None:
        """ Initializes the component class. """
        self.component_name: str = name
        self.inputs: List[ComponentInput] = []
        self.outputs: List[ComponentOutput] = []
        self.outputs_initialized: bool = False
        self.inputs_initialized: bool = False
        self.my_simulation_parameters: SimulationParameters = my_simulation_parameters
        if my_simulation_parameters is None:
            raise ValueError("My Simulation parameters was None.")
        self.simulation_repository: SimRepository
        self.default_connections: Dict[str, List[ComponentConnection]] = {}

    def add_default_connections(self, component: Type[Component], connections: List[ComponentConnection]) -> None:
        """ Adds a default connection list definition. """
        classname: str = component.get_classname()
        self.default_connections[classname] = connections
        log.trace("added connections: " + str(self.default_connections))

    def i_prepare_simulation(self) -> None:
        """ Gets called before the simulation to prepare the calculation. """
        raise NotImplementedError("Simulation preparation is missing for " + self.component_name + " (" + self.get_full_classname() + ")")

    def set_sim_repo(self, simulation_repository: SimRepository) -> None:
        """ Sets the SimRepository. """
        if simulation_repository is None:
            raise ValueError("simulation repository was none")
        self.simulation_repository = simulation_repository

    def add_input(self, object_name: str, field_name: str, load_type: lt.LoadTypes, unit: lt.Units,
                  mandatory: bool) -> ComponentInput:
        """ Adds an input definition. """
        myinput = ComponentInput(object_name, field_name, load_type, unit, mandatory)
        self.inputs.append(myinput)
        return myinput

    def add_output(self, object_name: str, field_name: str, load_type: lt.LoadTypes, unit: lt.Units,
                   postprocessing_flag: Optional[List[Any]] = None, sankey_flow_direction: Optional[bool] = None) -> ComponentOutput:
        """ Adds an output definition. """
        log.debug("adding output: " + field_name + " to component " + object_name)
        outp = ComponentOutput(object_name, field_name, load_type, unit, postprocessing_flag, sankey_flow_direction)
        self.outputs.append(outp)
        return outp

    def connect_input(self, input_fieldname: str, src_object_name: str, src_field_name: str) -> None:
        """ Connecting an input to an output. """
        if len(self.inputs) == 0:
            raise ValueError("The component " + self.component_name + " has no inputs.")
        component_input: ComponentInput
        input_to_set = None
        for component_input in self.inputs:
            if component_input.field_name == input_fieldname:
                if input_to_set is not None:
                    raise ValueError(
                        "The input " + input_fieldname + " of the component " + self.component_name + " was already set.")
                input_to_set = component_input
        if input_to_set is None:
            raise ValueError("The component " + self.component_name + " has no input with the name " + input_fieldname)
        input_to_set.src_object_name = src_object_name
        input_to_set.src_field_name = src_field_name

    def connect_dynamic_input(self, input_fieldname: str, src_object: ComponentOutput) -> None:
        """ For connecting an input to a dynamic output. """
        src_object_name = src_object.component_name
        src_field_name = src_object.field_name
        self.connect_input(input_fieldname=input_fieldname, src_object_name=src_object_name, src_field_name=src_field_name)

    # added variable input length and loop to be able to set default connections in one line in examples
    def connect_only_predefined_connections(self, *source_components):
        """ Wrapper for default connections and connect with connections list. """
        for source_component in source_components:
            connections = self.get_default_connections(source_component)
            self.connect_with_connections_list(connections)

    def connect_with_connections_list(self, connections: List[ComponentConnection]) -> None:
        """ Connect all inputs based on a connections list. """
        for connection in connections:
            src_name: str = typing.cast(str, connection.source_instance_name)
            self.connect_input(connection.target_input_name, src_name, connection.source_output_name)

    def get_default_connections(self, source_component: Type[Component]) -> List[ComponentConnection]:
        """ Gets the default connections for this component. """
        source_classname: str = source_component.get_classname()
        target_classname: str = self.get_classname()
        if source_classname not in self.default_connections:
            raise ValueError(
                "No default connections for " + source_classname + " in the connections for " + target_classname + ". content:\n" + str(
                    self.default_connections))
        connections = self.default_connections[source_classname]
        new_connections: List[ComponentConnection] = []
        for connection in connections:
            connection_copy = dc.replace(connection)
            connection_copy.source_instance_name = source_component.component_name
            new_connections.append(connection_copy)
        return new_connections

    def get_input_definitions(self) -> List[ComponentInput]:
        """ Gets the input definitions. """
        return self.inputs

    def get_outputs(self) -> List[ComponentOutput]:
        """ Delivers a list of outputs. """
        if len(self.outputs) == 0:
            raise Exception("Error: Component " + self.component_name + " has no outputs defined")
        return self.outputs

    def i_save_state(self) -> None:
        """ Abstract. Gets called at the beginning of a timestep to save the state. """
        raise NotImplementedError()

    def i_restore_state(self) -> None:
        """ Abstract. Restores the state of the component. Can be called many times while iterating. """
        raise NotImplementedError()

    def i_simulate(self, timestep: int, stsv: SingleTimeStepValues, force_convergence: bool) -> None:
        """ Performs the actual calculation. """
        raise NotImplementedError()

    def write_to_report(self) -> Any:
        """ Abstract function for writing the report entry for this component. """
        raise NotImplementedError("In " + self.component_name)

    def i_doublecheck(self, timestep: int, stsv: SingleTimeStepValues) -> None:
        """ Abstract. Gets called after the iterations are finished at each time step for potential debugging purposes. """
        pass  # noqa



# -*- coding: utf-8 -*-
# clean
""" Generic heating controller. """

from dataclasses import dataclass
# Owned
from typing import List, Any, Optional
# Generic/Built-in

from dataclasses_json import dataclass_json

from hisim.component import ConfigBase
from hisim import utils
from hisim import component as cp
from hisim import log
from hisim.components import generic_hot_water_storage_modular
from hisim.components import building
from hisim.components import controller_l2_energy_management_system
from hisim.loadtypes import LoadTypes, Units
from hisim.simulationparameters import SimulationParameters

__authors__ = "edited Johanna Ganglbauer"
__copyright__ = "Copyright 2021, the House Infrastructure Project"
__credits__ = ["Noah Pflugradt"]
__license__ = "MIT"
__version__ = "0.1"
__maintainer__ = "Vitor Hugo Bellotto Zago"
__email__ = "vitor.zago@rwth-aachen.de"
__status__ = "development"


@dataclass_json
@dataclass
class L1HeatPumpConfig(ConfigBase):

    """ L1 Controller Config. """

    name: str
    source_weight: int
    t_min_heating_in_celsius: float
    t_max_heating_in_celsius: float
    cooling_considered: bool
    t_min_cooling_in_celsius: Optional[float]
    t_max_cooling_in_celsius: Optional[float]
    day_of_heating_season_begin: Optional[int]
    day_of_heating_season_end: Optional[int]

    def __init__(self, name: str, source_weight: int, t_min_heating_in_celsius: float, t_max_heating_in_celsius: float, cooling_considered: bool,
                 t_min_cooling_in_celsius: Optional[float], t_max_cooling_in_celsius: Optional[float], day_of_heating_season_begin: Optional[int],
                 day_of_heating_season_end: Optional[int],  min_operation_time_in_seconds: int, min_idle_time_in_seconds: int):
        """ Initializes config. """
        # super.__init__(self)
        self.name = name
        self.source_weight = source_weight
        self.t_min_heating_in_celsius = t_min_heating_in_celsius
        self.t_max_heating_in_celsius = t_max_heating_in_celsius
        self.cooling_considered = cooling_considered
        self.t_min_cooling_in_celsius = t_min_cooling_in_celsius
        self.t_max_cooling_in_celsius = t_max_cooling_in_celsius
        self.day_of_heating_season_begin = day_of_heating_season_begin
        self.day_of_heating_season_end = day_of_heating_season_end
        self.min_operation_time_in_seconds = min_operation_time_in_seconds
        self.min_idle_time_in_seconds = min_idle_time_in_seconds

    @staticmethod
    def get_default_config_heat_source_controller(name: str) -> Any:
        """ Default Config for the buffer temperature. """
        config = L1HeatPumpConfig(name=name, source_weight=1, t_min_heating_in_celsius=23.0, t_max_heating_in_celsius=25.0,
                                  cooling_considered=True, t_min_cooling_in_celsius=23, t_max_cooling_in_celsius=25,
                                  day_of_heating_season_begin=270, day_of_heating_season_end=150, min_operation_time_in_seconds=1800, min_idle_time_in_seconds=1800)
        return config

    @staticmethod
    def get_default_config_heat_source_controller_buffer(name: str) -> Any:
        """ Default Config for the buffer temperature. """
        config = L1HeatPumpConfig(name=name, source_weight=1, t_min_heating_in_celsius=30.0, t_max_heating_in_celsius=50.0,
                                  cooling_considered=True, t_min_cooling_in_celsius=23, t_max_cooling_in_celsius=25,
                                  day_of_heating_season_begin=270, day_of_heating_season_end=150, min_operation_time_in_seconds=1800, min_idle_time_in_seconds=1800)
        return config

    @staticmethod
    def get_default_config_heat_source_controller_dhw(name: str) -> Any:
        """ Default Config for the buffer temperature. """
        config = L1HeatPumpConfig(name=name, source_weight=1, t_min_heating_in_celsius=40.0, t_max_heating_in_celsius=60.0,
                                  cooling_considered=False, t_min_cooling_in_celsius=23, t_max_cooling_in_celsius=25,
                                  day_of_heating_season_begin=270, day_of_heating_season_end=150, min_operation_time_in_seconds=1800, min_idle_time_in_seconds=1800)
        return config

class L1HeatPumpControllerState:

    """ Data class that saves the state of the controller. """

    def __init__(self, on_off: int, activation_time_step: int, deactivation_time_step: int, percentage: float) -> None:
        """ Initializes the heat pump controller state. """
        self.on_off: int = on_off
        self.activation_time_step: int = activation_time_step
        self.deactivation_time_step: int = deactivation_time_step
        self.percentage: float = percentage

    def clone(self) -> Any:
        """ Copies the current instance. """
        return L1HeatPumpControllerState(on_off=self.on_off, activation_time_step=self.activation_time_step,
                                         deactivation_time_step=self.deactivation_time_step,
                                         percentage=self.percentage)

    def i_prepare_simulation(self) -> None:
        """ Prepares the simulation. """
        pass

    def activate(self, timestep: int) -> None:
        """ Activates the heat pump and remembers the time step. """
        self.on_off = 1
        self.activation_time_step = timestep

    def deactivate(self, timestep: int) -> None:
        """ Deactivates the heat pump and remembers the time step. """
        self.on_off = 0
        self.deactivation_time_step = timestep


class L1HeatPumpController(cp.Component):

    """ L1 building controller. Processes signals ensuring comfort temperature of building.

    Gets available surplus electricity and the temperature of the storage or building to control as input,
    and outputs control signal 0/1 for turn off/switch on based on comfort temperature limits and available electricity.
    It optionally has different modes for cooling and heating selected by the time of the year.

    """

    # Inputs
    StorageTemperature = "StorageTemperature"
    StorageTemperatureModifier = "StorageTemperatureModifier"
    FlexibileElectricity = "FlexibleElectricity"
    # Outputs
    HeatControllerTargetPercentage = "HeatControllerTargetPercentage"
    OnOffState = "OnOffState"

    @utils.measure_execution_time
    def __init__(self, my_simulation_parameters: SimulationParameters, config: L1HeatPumpConfig) -> None:
        """ For initializing. """
        if not config.__class__.__name__ == L1HeatPumpConfig.__name__:
            raise ValueError("Wrong config class. Got a " + config.__class__.__name__)
        super().__init__(name=config.name + '_w' + str(config.source_weight), my_simulation_parameters=my_simulation_parameters)
        self.config: L1HeatPumpConfig = config
        self.minimum_runtime_in_timesteps = int(config.min_operation_time_in_seconds / self.my_simulation_parameters.seconds_per_timestep)
        self.minimum_resting_time_in_timesteps = int(config.min_idle_time_in_seconds / self.my_simulation_parameters.seconds_per_timestep)
        """ Initializes the class. """
        self.source_weight: int = config.source_weight
        self.cooling_considered: bool = config.cooling_considered
        if self.cooling_considered:
            if config.day_of_heating_season_begin is None:
                raise ValueError("Day of heating season begin was None")
            if config.day_of_heating_season_end is None:
                raise ValueError("Day of heating season end was None")
            self.heating_season_begin = config.day_of_heating_season_begin * 24 * 3600 / self.my_simulation_parameters.seconds_per_timestep
            self.heating_season_end = config.day_of_heating_season_end * 24 * 3600 / self.my_simulation_parameters.seconds_per_timestep
        self.state: L1HeatPumpControllerState = L1HeatPumpControllerState(0, 0, 0, 0)
        self.previous_state: L1HeatPumpConfig = self.state.clone()
        self.processed_state: L1HeatPumpConfig = self.state.clone()

        # Component Outputs
        self.heat_pump_target_percentage_channel: cp.ComponentOutput = self.add_output(self.component_name, self.HeatControllerTargetPercentage,
                                                                                       LoadTypes.ANY, Units.PERCENT)
        self.on_off_channel : cp.ComponentOutput = self.add_output(self.component_name, self.OnOffState,
                                                                                       LoadTypes.ANY, Units.ANY)

        # Component Inputs
        self.storage_temperature_channel: cp.ComponentInput = self.add_input(self.component_name, self.StorageTemperature, LoadTypes.TEMPERATURE,
                                                                             Units.CELSIUS, mandatory=True)
        self.storage_temperature_modifier_channel: cp.ComponentInput = self.add_input(self.component_name, self.StorageTemperatureModifier, LoadTypes.TEMPERATURE,
                                                                                      Units.CELSIUS, mandatory=False)

        self.flexible_electricity_input: cp.ComponentInput = self.add_input(self.component_name, self.FlexibileElectricity,
                                                                            LoadTypes.ELECTRICITY, Units.WATT, mandatory=False)

        self.add_default_connections(generic_hot_water_storage_modular.HotWaterStorage, self.get_buffer_default_connections())
        self.add_default_connections(building.Building, self.get_building_default_connections())
        self.add_default_connections(controller_l2_energy_management_system.L2GenericEnergyManagementSystem, self.get_ems_default_connections())

    def get_ems_default_connections(self):
        """ Sets the default connections for the building. """
        log.information("setting building default connections in L1 building Controller")
        connections = []
        ems_classname = controller_l2_energy_management_system.L2GenericEnergyManagementSystem.get_classname()
        connections.append(cp.ComponentConnection(L1HeatPumpController.FlexibileElectricity, ems_classname,
                                                  controller_l2_energy_management_system.L2GenericEnergyManagementSystem.FlexibleElectricity))
        return connections

    def get_buffer_default_connections(self):
        """ Sets default connections for the boiler. """
        log.information("setting buffer default connections in L1 building Controller")
        connections = []
        boiler_classname = generic_hot_water_storage_modular.HotWaterStorage.get_classname()
        connections.append(cp.ComponentConnection(L1HeatPumpController.StorageTemperature, boiler_classname,
                                                  generic_hot_water_storage_modular.HotWaterStorage.TemperatureMean))
        return connections

    def get_building_default_connections(self):
        """ Sets default connections for the boiler. """
        log.information("setting buffer default connections in L1 building Controller")
        connections = []
        building_classname = building.Building.get_classname()
        connections.append(cp.ComponentConnection(L1HeatPumpController.StorageTemperature, building_classname,
                                                  building.Building.TemperatureMean))
        return connections


    def i_prepare_simulation(self) -> None:
        """ Prepares the simulation. """
        pass

    def i_save_state(self) -> None:
        """ Saves the state. """
        self.previous_state = self.state.clone()

    def i_restore_state(self) -> None:
        """ Restores previous state. """
        self.state = self.previous_state.clone()

    def i_doublecheck(self, timestep: int, stsv: cp.SingleTimeStepValues) -> None:
        """ For double checking results. """
        pass

    def i_simulate(self, timestep: int, stsv: cp.SingleTimeStepValues, force_convergence: bool) -> None:
        """ Core Simulation function. """
        if force_convergence:
            # states are saved after each timestep, outputs after each iteration
            # outputs have to be in line with states, so if convergence is forced outputs are aligned to last known state.
            self.state = self.processed_state.clone()
        else:
            self.calculate_state(timestep, stsv)
            self.processed_state = self.state.clone()
        modulating_signal = self.state.percentage * self.state.on_off
        stsv.set_output_value(self.heat_pump_target_percentage_channel, modulating_signal)
        stsv.set_output_value(self.on_off_channel, self.state.on_off)

    def calc_percentage(self, t_storage: float, temperature_modifier: float) -> None:
        """ Calculate the heat pump target percentage. """
        if t_storage < self.config.t_min_heating_in_celsius:
            self.state.percentage = 1
            return
        if t_storage < self.config.t_max_heating_in_celsius and temperature_modifier == 0:
            self.state.percentage = 0.75
            return
        if t_storage >= self.config.t_max_heating_in_celsius and temperature_modifier == 0:
            self.state.percentage = 0.5
            return
        t_max_target = self.config.t_min_heating_in_celsius + temperature_modifier
        if t_storage < t_max_target:
            self.state.percentage = 0.5
            return

    def calculate_state(self, timestep: int, stsv: cp.SingleTimeStepValues) -> None:
        """ Calculate the heat pump state and activate / deactives. """
        t_storage = stsv.get_input_value(self.storage_temperature_channel)
        temperature_modifier = stsv.get_input_value(self.storage_temperature_modifier_channel)
        # return device on if minimum operation time is not fulfilled and device was on in previous state
        if self.state.on_off == 1 and self.state.activation_time_step + self.minimum_runtime_in_timesteps >= timestep:
            # mandatory on, minimum runtime not reached
            self.calc_percentage(t_storage, temperature_modifier)
            return
        if self.state.on_off == 0 and self.state.deactivation_time_step + self.minimum_resting_time_in_timesteps >= timestep:
            # mandatory off, minimum resting time not reached
            self.calc_percentage(t_storage, temperature_modifier)
            return
        # check signals and turn on or off if it is necessary
        t_min_target = self.config.t_min_heating_in_celsius + temperature_modifier
        # prevent heating in summer
        if self.cooling_considered:
            if self.heating_season_begin > timestep > self.heating_season_end and t_storage >= t_min_target - 5:
                self.state.deactivate(timestep)
                return
        if t_storage < t_min_target:
            self.state.activate(timestep)
            self.calc_percentage(t_storage, temperature_modifier)
            return
        t_max_target = self.config.t_max_heating_in_celsius + temperature_modifier
        if t_storage > t_max_target:
            self.calc_percentage(t_storage, temperature_modifier)
            self.state.deactivate(timestep)
            return

    def write_to_report(self) -> List[str]:
        """ Writes the information of the current component to the report. """
        lines: List[str] = []
        lines.append(f"Name: {self.component_name + str(self.config.source_weight)}")
        lines.append(self.config.get_string_dict())  # type: ignore
        return lines



""" Executes a json file in the simulator. """
# clean
import inspect
import os
from typing import List, Any, Dict
from inspect import isclass
from pkgutil import iter_modules
from pathlib import Path as PathlibPath
from importlib import import_module
from dataclasses import dataclass

from hisim.simulator import Simulator
from hisim import log
from hisim.json_generator import ConfigFile, ComponentEntry
from hisim.simulationparameters import SimulationParameters
from hisim import component as cp


@dataclass()
class ClassEntry:

    """ Entry for a single class. """

    file_name: str
    module_name: str
    class_name: str
    module: Any


class JsonExecutor:

    """ Class to read a json file created by the json generator and executre it. """

    def __init__(self, filename):
        """ Initializes the JsonExecutor. """
        with open(filename, "r", encoding="utf-8") as filestream:
            json_str = filestream.read()
            self.my_obj: ConfigFile = ConfigFile.from_json(json_str)
            if self.my_obj.my_simulation_parameters is None:
                raise ValueError("No simulation parameters object was found.")
            self.my_simulation_parameters: SimulationParameters = self.my_obj.my_simulation_parameters

    def component_importer(self, needed_classes: List[str]) -> Dict[str, Any]:
        """ Imports a config. """
        package_dir = os.path.join(PathlibPath(__file__).resolve().parent, "components")
        class_dict = {}
        for (_idx1, module_name, _idx2) in iter_modules([package_dir]):

            # import the module and iterate through its attributes
            module = import_module(f"hisim.components.{module_name}")

            for attribute_name in dir(module):
                attribute = getattr(module, attribute_name)
                if not isclass(attribute):
                    continue
                if attribute.__module__ != module.__name__:
                    # skipping because it's an import
                    continue

                # only import things that are actually needed
                full_class_name = attribute.__module__ + "." + attribute.__name__
                if full_class_name in needed_classes:
                    # Add the class to this package's variables
                    globals()[attribute_name] = attribute
                    class_dict[full_class_name] = attribute
        return class_dict

    def execute_all(self) -> None:
        """ Executes the json and starts the simulation. """
        needed_classes = self.generate_list_of_needed_classes_for_import()
        class_dict = self.component_importer(needed_classes)

        my_simulation_parameters: SimulationParameters = self.my_simulation_parameters

        simulator: Simulator = Simulator("json", "json_func", my_simulation_parameters)
        component_dict = {}
        for component_entry in self.my_obj.component_entries:
            component_instance = self.process_one_component_entry(component_entry, class_dict, my_simulation_parameters)
            simulator.add_component(component_instance)
            component_dict[component_entry.component_name] = component_instance
        for component_entry in self.my_obj.component_entries:
            if len(component_entry.default_connections) == 0:
                continue
            for component_name in component_entry.default_connections:
                dst_class: cp.Component = component_dict[component_entry.component_name]
                src_class = component_dict[component_name]
                dst_class.connect_only_predefined_connections(src_class)
        simulator.run_all_timesteps()

    def process_one_component_entry(self, component_entry: ComponentEntry, class_dict: Any, my_simulation_parameters: SimulationParameters) -> Any:
        """ Processes a single component entry in the json. """
        my_class = class_dict[component_entry.component_full_classname]
        signature_component = inspect.signature(my_class)
        found_simulation_parameters = False
        found_config = False
        for parameter_name in signature_component.parameters:
            if parameter_name == "my_simulation_parameters":
                found_simulation_parameters = True
            elif parameter_name == "config":
                found_config = True
            else:
                log.warning("Found unclear parameter: " + parameter_name + " in component " + component_entry.component_full_classname)
        if not found_simulation_parameters:
            raise ValueError("The class " + component_entry.component_full_classname + " has no simulation parameters parameter")
        if not found_config:
            raise ValueError("The class " + component_entry.component_full_classname + " has no config parameter")
        config_class = class_dict[component_entry.config_full_classname]

        config_instance = config_class.from_dict(component_entry.configuration)

        class_instance = my_class(config=config_instance, my_simulation_parameters=my_simulation_parameters)
        return class_instance

    def generate_list_of_needed_classes_for_import(self):
        """ Generates a list of classes that need to be important from the object list. """
        needed_classes: List[str] = []
        for component_entry in self.my_obj.component_entries:
            needed_key = component_entry.component_full_classname
            needed_classes.append(needed_key)
            needed_config_key = component_entry.config_full_classname
            needed_classes.append(needed_config_key)
        return needed_classes



class WarmWaterStorageConfig:
    tank_diameter = 1 # 0.9534        # [m]
    tank_height = 2 # 3.15              # [m]
    tank_start_temperature: float = 65  # [°C]
    temperature_difference = 0.3        # [°C]
    tank_u_value =  0 # 0.35                 # [W/m^2*K]
    slice_height_minimum = 0.05         # [m]

class CHPControllerConfig:
    """
    The CHP controller is used to implement an on and off hysteresis
    Decide if its heat- or electricity-led

    Two temperature sensors in the tank are giving the needed information. They can be set at a height percentage in the tank. =% is the top, 100 % s the bottom of the tank.
    If the T at the upper sensor is below temperature_switch_on the chp will run until the lower sensor is above temperature_switch_off.
    A minimum runtime in minutes can be defined for the chp.

    If the chp is electric-led, ths is not needed and the electricity demand is provided directly to the chp
    """
    method_of_operation = "heat"
    temperature_switch_on = 60          # [°C]
    temperature_switch_off = 65         # [°C]

    # in steps of 20 % [0, 20, 40 ,60, 80, 100]
    heights_in_tank = [0, 20, 40, 60, 80, 100]
    height_upper_sensor = 20            # [%]
    height_lower_sensor = 60            # [%]

    minimum_runtime_minutes = 4000        # [min]

class GasHeaterConfig:
    is_modulating = True
    P_th_min = 1_000                    # [W]
    P_th_max = 12_000                    # [W]
    eff_th_min = 0.60                   # [-]
    eff_th_max = 0.90                   # [-]
    delta_T = 25
    mass_flow_max = P_th_max / (4180 * delta_T)     # kg/s ## -> ~0.07
    temperature_max = 80                # [°C]

class GasControllerConfig:
    """
    This controller works like the CHP controller, but switches on later so the CHP is used more often.
    Gas heater is used as a backup if the CHP power is not high enough.
    If the minimum_runtime is smaller than the timestep, the minimum_runtime is 1 timestep --> generic_gas_heater.py
    """
    temperature_switch_on = 55          # [°C]
    temperature_switch_off = 70         # [°C]

    # in steps of 20 % [0, 20, 40 ,60, 80, 100]
    height_upper_sensor = 20            # [%]
    height_lower_sensor = 80            # [%]

    # minimal timestep is minute
    minimum_runtime_minutes = 7000        # [min]

class LoadConfig:
    # massflow_load_minute = 2.5          # [kg/min]
    # massflow_load = massflow_load_minute / 60   # [kg/s]

    possible_massflows_load = [0.1, 0.2, 0.3, 0.4]              # [kg/s]
    delta_T = 20

    # the returnflow shows if there was enough energy in the water
    # -> use in load! Not in storage, there the water from WW is included
    temperature_returnflow_minimum = 30             # [°C]

    kWh_per_year = 20_201
    demand_factor = kWh_per_year / 1000

class ElectricityDemandConfig:
    kWh_per_year = 6000
    demand_factor = kWh_per_year / 1000

class HouseholdWarmWaterDemandConfig:
    freshwater_temperature = 10         # [°C]
    ww_temperature_demand = 45          # [°C]
    # german --> Grädigkeit
    # difference between T1_in and T2_out
    temperature_difference_hot = 5      # [°C]
    temperature_difference_cold = 6     # [°C]

    heat_exchanger_losses = 0

    kWh_per_year = 2000
    demand_factor = kWh_per_year / 1000

class HydrogenStorageConfig:
    # combination of
    min_capacity = 0                    # [kg_H2]
    max_capacity = 500                  # [kg_H2]

    starting_fill = 400                 # [kg_H2]

    max_charging_rate_hour = 2          # [kg/h]
    max_discharging_rate_hour = 2       # [kg/h]
    max_charging_rate = max_charging_rate_hour / 3600
    max_discharging_rate = max_discharging_rate_hour / 3600

    # ToDo: How does the necessary Heat/Energy come to the Storage?
    energy_for_charge = 0               # [kWh/kg]
    energy_for_discharge = 0            # [kWh/kg]

    loss_factor_per_day = 0             # [lost_%/day]

class AdvElectrolyzerConfig:
    waste_energy = 400                    # [W]   # 400
    min_power = 1_400                     # [W]   # 1400
    max_power = 2_4000                   # [W]   # 2400
    min_power_percent = 60            # [%]
    max_power_percent = 100             # [%]
    min_hydrogen_production_rate_hour = 300  # [Nl/h]
    max_hydrogen_production_rate_hour = 5000   # [Nl/h]   #500
    min_hydrogen_production_rate = min_hydrogen_production_rate_hour / 3600  # [Nl/s]
    max_hydrogen_production_rate = max_hydrogen_production_rate_hour / 3600  # [Nl/s]
    pressure_hydrogen_output = 30       # [bar]     --> max pressure mode at 35 bar

    """
    The production rate can be converted to an efficiency.
    eff_electrolyzer = (production_rate_hour * hydrogen_specific_heat_capacity_per_kg[kWh/kg]) / (Power_this_timestep[kWh] * hydrogen_specific_volume [m³kg])

    in the component electrolyzer:
    hydrogen_output = Power_this_timestep[kWh] * eff_electrolyzer / hydrogen_specific_heat_capacity_per_kg[kWh/kg]

    I think its overengineering because the providers give the needed information and we try to calculate it back and forth

    --> Solution: efficiency of the electrolyzer is calculated and is an Output
    """

class PVConfig:
    peak_power = 20_000                 # [W]

class ExtendedControllerConfig:
    # Active Components
    chp = True
    gas_heater = True
    electrolyzer = True
    # electrolyzer = False

    #power mode chp
    # chp_mode = "heat"
    chp_mode = "power"
    chp_power_states_possible = 10
    maximum_autarky = False

class PhysicsConfig:
    water_density = 1000                                        # [kg/m^3]
    water_specific_heat_capacity_in_joule_per_kilogram_per_kelvin = 4_180                         # J/kgK

    # Schmidt 2020: Wasserstofftechnik  S.170ff
    # fuel value H2:    10.782 MJ/m³    (S.172)
    # density H2:       0.08989 kg/m³   (S. 23) -> standard conditions
    hydrogen_density = 0.08989                                  # [kg/m³]
    hydrogen_specific_volume = 1 / hydrogen_density             # [m^3/kg]
    hydrogen_specific_fuel_value_per_m_3 = 10.782 * 10 ** 6     # [J/m³]
    hydrogen_specific_fuel_value_per_kg = hydrogen_specific_fuel_value_per_m_3 / hydrogen_density   # [J/kg]

    # Schmidt 2020: Wasserstofftechnik  S.170ff
    # fuel value Methan:    35.894 MJ/m³    (S.172)
    # density Methan:       0.71750 kg/m³   (S. 23) -> standard conditions
    natural_gas_density = 0.71750                                  # [kg/m³]
    natural_gas_specific_volume = 1 / hydrogen_density             # [m^3/kg]
    natural_gas_specific_fuel_value_per_m_3 = 35.894 * 10 ** 6     # [J/m³]
    natural_gas_specific_fuel_value_per_kg = natural_gas_specific_fuel_value_per_m_3 / natural_gas_density   # [J/kg]



# Generic/Built-in
import numpy as np
import copy
import matplotlib
import seaborn
from math import pi
from dataclasses import dataclass
from dataclasses_json import dataclass_json
from typing import Optional, List, Any

# Owned
import hisim.utils as utils
import hisim.loadtypes as lt
from hisim import component as cp
from hisim.simulationparameters import SimulationParameters
from hisim.components.controller_l2_energy_management_system import L2GenericEnergyManagementSystem
from hisim.components.weather import Weather
from hisim.components import controller_l1_heatpump
from hisim import log

__authors__ = "edited Johanna Ganglbauer"
__copyright__ = "Copyright 2021, the House Infrastructure Project"
__credits__ = ["Noah Pflugradt"]
__license__ = "MIT"
__version__ = "0.1"
__maintainer__ = "Vitor Hugo Bellotto Zago"
__email__ = "vitor.zago@rwth-aachen.de"
__status__ = "development"


@dataclass_json
@dataclass
class HeatPumpConfig:
    name: str
    source_weight: int
    parameter_string: str
    manufacturer: str
    device_name: str
    power_th: float
    cooling_considered: bool
    heating_season_begin: Optional[int]
    heating_season_end: Optional[int]

    def __init__(self, name: str, source_weight: int, manufacturer: str, device_name: str, power_th: float, cooling_considered: bool,
                 heating_season_begin: Optional[int], heating_season_end: Optional[int]):
        self.name = name
        self.source_weight = source_weight
        self.manufacturer = manufacturer
        self.device_name = device_name
        self.power_th = power_th
        self.cooling_considered = cooling_considered
        self.heating_season_begin = heating_season_begin
        self.heating_season_end = heating_season_end
        
    @staticmethod
    def get_default_config_heating() -> Any:
        config = HeatPumpConfig(name='HeatingHeatPump', source_weight=1, manufacturer="Viessmann Werke GmbH & Co KG",
                                device_name="Vitocal 300-A AWO-AC 301.B07", power_th=6200, cooling_considered=True, heating_season_begin=270,
                                heating_season_end=150)
        return config

    @staticmethod
    def get_default_config_waterheating() -> Any:
        config = HeatPumpConfig(name='DHWHeatPump', source_weight=1, manufacturer="Viessmann Werke GmbH & Co KG",
                                device_name="Vitocal 300-A AWO-AC 301.B07", power_th=3000, cooling_considered=False, heating_season_begin=None,
                                heating_season_end=None)
        return config

    @staticmethod
    def get_default_config_heating_electric() -> Any:
        config = HeatPumpConfig(name='HeatingHeatingRod', source_weight=1, manufacturer="dummy", device_name="HeatingRod", power_th=6200,
                                cooling_considered=False, heating_season_begin=None, heating_season_end=None)
        return config

    @staticmethod
    def get_default_config_waterheating_electric() -> Any:
        config = HeatPumpConfig(name='DHWHeatingRod', source_weight=1, manufacturer="dummy", device_name="HeatingRod", power_th=3000,
                                cooling_considered=False, heating_season_begin=None, heating_season_end=None)
        return config


class ModularHeatPumpState:
    """
    This data class saves the state of the heat pump.
    """

    def __init__(self, state: float = 0, timestep: int = -1):
        self.state: float = state
        self.timestep = timestep

    def clone(self):
        return ModularHeatPumpState(state=self.state, timestep=self.timestep)


class ModularHeatPump(cp.Component):
    """
    Heat pump implementation. It does support a refrigeration cycle. The heatpump_modular differs to heatpump in (a) minumum run- and
    idle time are given in seconds (not in time steps), (b) the season for heating and cooling is explicitly separated by days of the year.
    This is mostly done to avoid heating and cooling at the same day in spring and autum with PV surplus available. (c) heat pump modular needs
    a generic_controller_l1_runtime signal. The run time is not controlled in the component itself but in the controller.
    
    STILL TO BE DONE: implement COP for cooling period. At the moment it cools with heating efficiencies.

    Parameters
    ----------
    manufacturer : str
        Heat pump manufacturer
    device_name : str
        Heat pump model
    heating_season_begin : int, optional
        Day( julian day, number of day in year ), when heating season starts - and cooling season ends. The default is 270.
    heating_season_end : int, optional
        Day( julian day, number of day in year ), when heating season ends - and cooling season starts. The default is 150
    name : str, optional
        Name of heatpump within simulation. The default is 'HeatPump'
    source_weight : int, optional
        Weight of component, relevant if there is more than one heat pump, defines hierachy in control. The default is 1.
        
    """
    # Inputs
    TemperatureOutside = "TemperatureOutside"
    L1DeviceSignal = "L1DeviceSignal"
    l1_RunTimeSignal = 'l1_RunTimeSignal'
    ems_flexible_electricity = "EMS Modulating Signal"

    # Outputs
    ThermalPowerDelivered = "ThermalPowerDelivered"
    ElectricityOutput = "ElectricityOutput"
    PowerModifier = "PowerModifier"

    # Similar components to connect to:
    # 1. HeatPump l1 controller
    # 2. HeatPump l2 controller
    # 3. HeatPump l3 controller ( optional )

    @utils.measure_execution_time
    def __init__(self, config: HeatPumpConfig, my_simulation_parameters: SimulationParameters):

        super().__init__(name=config.name + '_w' + str(config.source_weight), my_simulation_parameters=my_simulation_parameters)
        self.config = config
        self.build(config)

        if my_simulation_parameters.system_config.clever:
            postprocessing_flag = [lt.InandOutputType.ELECTRICITY_CONSUMPTION_EMS_CONTROLLED]
        else:
            postprocessing_flag = [lt.InandOutputType.ELECTRICITY_CONSUMPTION_UNCONTROLLED]

        # Inputs - Mandatories
        self.TemperatureOutsideC: cp.ComponentInput = self.add_input(self.component_name, self.TemperatureOutside, lt.LoadTypes.ANY, lt.Units.CELSIUS,
                                                                     mandatory=True)

        self.EMS_Flexible_ElectricityC: cp.ComponentInput = self.add_input(self.component_name, self.ems_flexible_electricity,
                                                                           lt.LoadTypes.ELECTRICITY, lt.Units.WATT, mandatory=False)

        self.L1HeatControllerTargetPercentage: cp.ComponentInput = self.add_input(self.component_name, self.L1DeviceSignal, lt.LoadTypes.ANY, lt.Units.PERCENT,
                                                                 mandatory=True)

        # Outputs
        self.ThermalPowerDeliveredC: cp.ComponentOutput = self.add_output(object_name=self.component_name, field_name=self.ThermalPowerDelivered,
                                                                          load_type=lt.LoadTypes.HEATING, unit=lt.Units.WATT,
                                                                          postprocessing_flag=[lt.InandOutputType.HEAT_TO_BUFFER])
        self.ElectricityOutputC: cp.ComponentOutput = self.add_output(object_name=self.component_name, field_name=self.ElectricityOutput,
                                                                      load_type=lt.LoadTypes.ELECTRICITY, unit=lt.Units.WATT,
                                                                      postprocessing_flag=postprocessing_flag)

        self.PowerModifierChannel: cp.ComponentOutput = self.add_output(object_name=self.component_name, field_name=self.PowerModifier,
                                                                      load_type=lt.LoadTypes.ANY, unit=lt.Units.ANY,postprocessing_flag=[])

        self.add_default_connections(Weather, self.get_weather_default_connections())
        self.add_default_connections(controller_l1_heatpump.L1HeatPumpController, self.get_l1_heatpump_controller_default_connections())
        self.add_default_connections(L2GenericEnergyManagementSystem, self.get_ems_default_connections())

    def get_weather_default_connections(self):
        log.information("setting weather default connections in HeatPump")
        connections = []
        weather_classname = Weather.get_classname()
        connections.append(cp.ComponentConnection(ModularHeatPump.TemperatureOutside, weather_classname, Weather.TemperatureOutside))
        return connections

    def get_ems_default_connections(self):
        log.information("setting weather default connections in HeatPump")
        connections = []
        ems_classname = L2GenericEnergyManagementSystem.get_classname()
        connections.append(
            cp.ComponentConnection(ModularHeatPump.ems_flexible_electricity, ems_classname, L2GenericEnergyManagementSystem.FlexibleElectricity))
        return connections

    def get_l1_heatpump_controller_default_connections(self):
        log.information("setting l1 default connections in HeatPump")
        connections = []
        controller_classname = controller_l1_heatpump.L1HeatPumpController.get_classname()
        connections.append(cp.ComponentConnection(ModularHeatPump.L1DeviceSignal, controller_classname,
                                                  controller_l1_heatpump.L1HeatPumpController.HeatControllerTargetPercentage))
        return connections

    def i_prepare_simulation(self) -> None:
        """ Prepares the simulation. """
        pass

    def build(self, config):
        self.name = config.name
        self.source_weight = config.source_weight
        self.manufacturer = config.manufacturer
        self.devicename = config.device_name

        # Retrieves heat pump from database - BEGIN
        heat_pumps_database = utils.load_smart_appliance("Heat Pump")

        heat_pump_found = False
        for heat_pump in heat_pumps_database:
            if heat_pump["Manufacturer"] == config.manufacturer and heat_pump["Name"] == config.device_name:
                heat_pump_found = True
                break

        if heat_pump_found == False:
            raise Exception("Heat pump model not registered in the database")

        # Interpolates COP data from the database
        self.cop_ref = []
        self.t_out_ref = []
        for heat_pump_cops in heat_pump['COP']:
            self.t_out_ref.append(float([*heat_pump_cops][0][1:].split("/")[0]))
            self.cop_ref.append(float([*heat_pump_cops.values()][0]))
        self.cop_coef = np.polyfit(self.t_out_ref, self.cop_ref, 1)
        self.power_th = config.power_th
        self.cooling_considered = config.cooling_considered
        if self.cooling_considered:
            self.heating_season_begin = config.heating_season_begin * 24 * 3600 / self.my_simulation_parameters.seconds_per_timestep
            self.heating_season_end = config.heating_season_end * 24 * 3600 / self.my_simulation_parameters.seconds_per_timestep

        self.state = ModularHeatPumpState()
        self.previous_state = ModularHeatPumpState()

        # Writes info to report
        self.write_to_report()

    def cal_cop(self, t_out: float) -> float:
        val: float = self.cop_coef[0] * t_out + self.cop_coef[1]
        return val

    def i_save_state(self) -> None:
        self.previous_state = self.state.clone()

    def i_restore_state(self) -> None:
        self.state = self.previous_state.clone()

    def i_doublecheck(self, timestep: int, stsv: cp.SingleTimeStepValues) -> None:
        pass

    def write_to_report(self) -> List[str]:
        lines: List[str] = []
        lines.append("Name: {}".format(self.name + str(self.source_weight)))
        lines.append("Manufacturer: {}".format(self.name))
        lines.append("Max power: {:4.0f} kW".format((self.power_th) * 1E-3))
        return lines

    def i_simulate(self, timestep: int, stsv: cp.SingleTimeStepValues, force_convergence: bool) -> None:

        # Inputs
        target_percentage = stsv.get_input_value(self.L1HeatControllerTargetPercentage)

        T_outside: float = stsv.get_input_value(self.TemperatureOutsideC)
        cop = self.cal_cop(T_outside)
        electric_power = self.power_th / cop

        # calculate modulation
        if target_percentage > 0:
            power_modifier = target_percentage
        if target_percentage == 0:
            power_modifier = 0
        if target_percentage < 0:
            flexible_electricity = stsv.get_input_value(self.EMS_Flexible_ElectricityC)
            power_modifier = flexible_electricity / electric_power

        if power_modifier > 1:
            power_modifier = 1

        stsv.set_output_value(self.ThermalPowerDeliveredC, self.power_th * power_modifier)
        stsv.set_output_value(self.PowerModifierChannel, power_modifier)

        stsv.set_output_value(self.ElectricityOutputC,  electric_power * power_modifier)

        # #put forecast into dictionary  # if self.my_simulation_parameters.system_config.predictive:  #     #only in first timestep  #     if self.state.timestep + 1 == timestep:  #         self.state.timestep += 1  #         self.previous_state.timestep += 1  #         runtime = stsv.get_input_value( self.l1_RunTimeSignalC )

        #         self.simulation_repository.set_dynamic_entry(component_type = lt.ComponentType.HEAT_PUMP, source_weight = self.source_weight, entry =[self.power_th / cop] * runtime)




""" Generates a Json file for the json executor. """
# clean
from typing import Any, Optional, List, Type, Dict
from dataclasses import dataclass, field
from dataclass_wizard import JSONWizard

from hisim import log
from hisim.component import ConfigBase
from hisim.simulationparameters import SimulationParameters


@dataclass
class ConnectionEntry(JSONWizard):

    """ Represents a connection between two components. """

    src_component_name: str
    src_output_name: str
    dst_input_name: str


@dataclass
class ComponentEntry(JSONWizard):

    """ Entry for a single component. """

    component_full_classname: str
    component_name: str
    configuration: Dict[Any, Any]
    config_full_classname: str
    default_connections: List[str]
    manual_connections: List[ConnectionEntry]


@dataclass
class ConfigFile(JSONWizard):

    """ Contains all the data that goes into the json config file. """

    system_name: str
    component_entries: List[ComponentEntry] = field(default_factory=list)
    my_simulation_parameters: Optional[SimulationParameters] = None


class JsonConfigurationGenerator:

    """ Class to generate Json config files that can be used to start calculations without Python. """

    def __init__(self, name: str) -> None:
        """ Initializes the configuration generator. """
        self.config_file: ConfigFile = ConfigFile(system_name=name)

    def set_simulation_parameters(self, my_simulation_parameter: SimulationParameters) -> None:
        """ Sets the simulation parameters. """
        self.config_file.my_simulation_parameters = my_simulation_parameter

    def add_component(self, config: Type[ConfigBase]) -> ComponentEntry:
        """ Adds a component and returns a component entry. """
        config_json_str = config.to_dict()
        component_entry = ComponentEntry(component_full_classname=config.get_main_classname(),
                                         component_name=config.name,
                                         configuration=config_json_str,
                                         config_full_classname=config.get_config_classname(),
                                         default_connections=[],
                                         manual_connections=[])
        self.config_file.component_entries.append(component_entry)
        log.information("Added component " + config.name)
        return component_entry

    def add_default_connection(self, from_entry: ComponentEntry, to_entry: ComponentEntry) -> None:
        """ Adds a default connection to another component. """
        to_entry.default_connections.append(from_entry.component_name)

    def add_manual_connection(self, from_entry: ComponentEntry, output_name: str, to_entry: ComponentEntry, input_name: str) -> None:
        """ Adds a manual connection entry for connecting a single input. """
        my_connection = ConnectionEntry(from_entry.component_name, output_name, input_name)
        to_entry.manual_connections.append(my_connection)

    def save_to_json(self, filename: str) -> None:
        """ Saves a configuration to a json file. """
        with open(filename, "w", encoding="utf-8") as filestream:
            mystr = self.config_file.to_json(indent=4)
            filestream.write(mystr)



from typing import List, Any
from hisim import component as cp
from hisim import loadtypes as lt
from hisim.components.configuration import PhysicsConfig
from hisim import utils
#from math import pi
#from math import floor
from hisim.simulationparameters import SimulationParameters
from hisim.components import controller_l2_generic_heat_simple
from hisim.components import generic_hydrogen_storage
import hisim.log as log
import pandas as pd
import os
from dataclasses import dataclass
from dataclasses_json import dataclass_json
import math

__authors__ = "Frank Burkrad, Maximilian Hillen,"
__copyright__ = "Copyright 2021, the House Infrastructure Project"
__credits__ = ["Noah Pflugradt"]
__license__ = ""
__version__ = ""
__maintainer__ = "Johanna Ganglbauer"
__email__ = "johanna.ganglbauer@4wardenergy.at"
__status__ = "development"

@dataclass_json
@dataclass
class GCHPConfig:
    """
    GCHP Config
    """
    name: str
    source_weight: int
    p_el : float       
    p_th : float
    p_fuel : float

    def __init__( self,
                  name : str,
                  source_weight : int,
                  p_el : float,
                  p_th : float,
                  p_fuel : float ):
        self.name = name
        self.source_weight = source_weight
        self.p_el = p_el
        self.p_th = p_th
        self.p_fuel = p_fuel
        
    @staticmethod
    def get_default_config() -> Any:
        config=GCHPConfig(name='CHP',
                          source_weight=1,
                          p_el=2000,
                          p_th=3000,
                          p_fuel=6000) 
        return config
            
class GenericCHPState:
    """
    This data class saves the state of the CHP.
    """

    def __init__( self, state : float = 0 ) -> None:
        self.state:float = state
        
    def clone( self ) -> Any:
        return GenericCHPState(state = self.state)
            
class GCHP( cp.Component ):
    """
    Simulates CHP operation with constant electical and thermal power as well as constant fuel consumption.
    """
    # Inputs
    L1DeviceSignal = "L1DeviceSignal"

    # Outputs
    ThermalPowerDelivered = "ThermalPowerDelivered"
    ElectricityOutput = "ElectricityOutput"
    FuelDelivered = "FuelDelivered"
    
    def __init__( self, my_simulation_parameters: SimulationParameters, config: GCHPConfig ) -> None:
        super().__init__(name=config.name + '_w' + str(config.source_weight), my_simulation_parameters=my_simulation_parameters )
        self.build(config)

        #Inputs
        self.L1DeviceSignalC: cp.ComponentInput = self.add_input(self.component_name,
                                                                  self.L1DeviceSignal,
                                                                  lt.LoadTypes.ON_OFF,
                                                                  lt.Units.BINARY,
                                                                  mandatory = True)
        
        #Component outputs
        self.ThermalPowerDeliveredC: cp.ComponentOutput = self.add_output(
            object_name=self.component_name, field_name=self.ThermalPowerDelivered, load_type=lt.LoadTypes.HEATING,
            unit=lt.Units.WATT, postprocessing_flag=[lt.InandOutputType.THERMAL_PRODUCTION])
        self.ElectricityOutputC: cp.ComponentOutput = self.add_output(
            object_name=self.component_name, field_name=self.ElectricityOutput, load_type=lt.LoadTypes.ELECTRICITY,
            unit=lt.Units.WATT, postprocessing_flag=[lt.InandOutputType.ELECTRICITY_PRODUCTION, lt.ComponentType.FUEL_CELL])
        self.FuelDeliveredC: cp.ComponentOutput = self.add_output(self.component_name,
                                                                  self.FuelDelivered,
                                                                  lt.LoadTypes.HYDROGEN,
                                                                  lt.Units.KG_PER_SEC)
        self.add_default_connections(L1GenericCHPRuntimeController, self.get_l1_controller_default_connections())
    
    def build( self, config: GCHPConfig ) -> None:
        self.state = GenericCHPState()
        self.previous_state = GenericCHPState()
        self.name = config.name
        self.source_weight = config.source_weight
        self.p_th = config.p_th
        self.p_el = config.p_el
        self.p_fuel = config.p_fuel * 1e-8 / 1.41 #converted to kg / s
    def i_prepare_simulation(self) -> None:
        """ Prepares the simulation. """
        pass
    def i_save_state(self) -> None:
        self.previous_state = self.state.clone( )

    def i_restore_state(self) -> None:
        self.state = self.previous_state.clone( )

    def i_doublecheck(self, timestep: int, stsv: cp.SingleTimeStepValues) -> None:
         pass

    def i_simulate( self, timestep: int, stsv: cp.SingleTimeStepValues,  force_convergence: bool ) -> None:

        # Inputs
        self.state.state = stsv.get_input_value( self.L1DeviceSignalC )
       
        # Outputs
        stsv.set_output_value( self.ThermalPowerDeliveredC, self.state.state * self.p_th )
        stsv.set_output_value( self.ElectricityOutputC, self.state.state * self.p_el )
        
        #heat of combustion hydrogen: 141.8 MJ / kg; conversion W = J/s to kg / s
        stsv.set_output_value( self.FuelDeliveredC, self.state.state * self.p_fuel )
        
    def get_l1_controller_default_connections( self ) -> List[cp.ComponentConnection]:
        log.information("setting l1 default connections in generic CHP" )
        connections: List[cp.ComponentConnection] = [ ]
        controller_classname = L1GenericCHPRuntimeController.get_classname()
        connections.append(cp.ComponentConnection(GCHP.L1DeviceSignal, controller_classname, L1GenericCHPRuntimeController.L1DeviceSignal))
        return connections
    
    def write_to_report(self):
        lines = []
        lines.append("CHP operation with constant electical and thermal power: {}".format( self.name + str( self.source_weight ) ) )
        lines.append( "P_el {:4.0f} kW".format( self.p_el ) )
        lines.append( "P_th {:4.0f} kW".format( self.p_th ) )
        return lines
        
@dataclass_json
@dataclass
class L1CHPConfig:
    """
    L1CHP Config
    """
    name : str
    source_weight : int
    min_operation_time : int
    min_idle_time : int
    min_h2_soc : float

    def __init__( self,
                  name : str,
                  source_weight : int,
                  min_operation_time : int,
                  min_idle_time : int,
                  min_h2_soc : float ) -> None:
        self.name = name
        self.source_weight = source_weight
        self.min_operation_time = min_operation_time
        self.min_idle_time = min_idle_time
        self.min_h2_soc = min_h2_soc
        
    @staticmethod
    def get_default_config() -> Any:
        config = L1CHPConfig( name = 'L1CHPRunTimeController',
                              source_weight =  1,
                              min_operation_time = 14400,
                              min_idle_time = 7200,
                              min_h2_soc = 5 )
        return config
        
class L1GenericCHPControllerState:
    """
    This data class saves the state of the controller.
    """

    def __init__( self, timestep_actual : int = -1, state : int = 0, timestep_of_last_action : int = 0 ) -> None:
        self.timestep_actual = timestep_actual
        self.state = state
        self.timestep_of_last_action = timestep_of_last_action
        
    def clone( self ) -> Any:
        return L1GenericCHPControllerState(timestep_actual = self.timestep_actual, state = self.state, timestep_of_last_action = self.timestep_of_last_action)
    
    def is_first_iteration( self, timestep: int ) -> bool:
        if self.timestep_actual + 1 == timestep:
            self.timestep_actual += 1
            return True
        else:
            return False
        
    def activation( self, timestep: int ) -> None:
        self.state = 1
        self.timestep_of_last_action = timestep
        
    def deactivation( self, timestep: int ) -> None:
        self.state = 0
        self.timestep_of_last_action = timestep 

class L1GenericCHPRuntimeController(cp.Component):
    
    """
    L1 CHP Controller. It takes care of the operation of the CHP only in terms of running times.

    Parameters
    --------------
    min_running_time: int, optional
        Minimal running time of device, in seconds. The default is 3600 seconds.
    min_idle_time : int, optional
        Minimal off time of device, in seconds. The default is 900 seconds.
    source_weight : int, optional
        Weight of component, relevant if there is more than one component of same type, defines hierachy in control. The default is 1.
    component type : str, optional
        Name of component to be controlled
    """
    # Inputs
    l2_DeviceSignal = "l2_DeviceSignal"
    ElectricityTarget = "ElectricityTarget"
    HydrogenSOC = "HydrogenSOC"

    # Outputs
    L1DeviceSignal = "L1DeviceSignal"
    
    # Similar components to connect to:
    # 1. Building
    @utils.measure_execution_time
    def __init__( self, 
                  my_simulation_parameters : SimulationParameters,
                  config : L1CHPConfig ):
        
        super().__init__(name=config.name + '_w' + str(config.source_weight), 
                          my_simulation_parameters=my_simulation_parameters )
        
        self.build( config )
        
        # add inputs
        self.l2_DeviceSignalC: cp.ComponentInput = self.add_input(self.component_name,
                                                                  self.l2_DeviceSignal,
                                                                  lt.LoadTypes.ON_OFF,
                                                                  lt.Units.BINARY,
                                                                  mandatory = True)
        self.ElectricityTargetC : cp.ComponentInput = self.add_input(self.component_name,
                                                                     self.ElectricityTarget,
                                                                     lt.LoadTypes.ELECTRICITY,
                                                                     lt.Units.WATT,
                                                                     mandatory = True)
        
        self.HydrogenSOCC : cp.ComponentInput = self.add_input(self.component_name,
                                                               self.HydrogenSOC,
                                                               lt.LoadTypes.HYDROGEN,
                                                               lt.Units.PERCENT,
                                                               mandatory = True)

        self.add_default_connections(controller_l2_generic_heat_simple.L2GenericHeatController, self.get_l2_controller_default_connections())
        self.add_default_connections(generic_hydrogen_storage.GenericHydrogenStorage, self.get_hydrogen_storage_default_connections())
        
        
        # add outputs
        self.L1DeviceSignalC: cp.ComponentOutput = self.add_output(self.component_name,
                                                                    self.L1DeviceSignal,
                                                                    lt.LoadTypes.ON_OFF,
                                                                    lt.Units.BINARY)
        
    def get_l2_controller_default_connections( self ) -> List[cp.ComponentConnection]:
        log.information("setting l2 default connections in l1")
        connections: List[cp.ComponentConnection] = [ ]
        controller_classname = controller_l2_generic_heat_simple.L2GenericHeatController.get_classname()
        connections.append(cp.ComponentConnection(L1GenericCHPRuntimeController.l2_DeviceSignal, controller_classname, controller_l2_generic_heat_simple.L2GenericHeatController.l2_device_signal))
        return connections
    
    def get_hydrogen_storage_default_connections( self ) -> List[cp.ComponentConnection]:
        log.information("setting generic H2 storage default connections in L1 of generic CHP" )
        connections: List[cp.ComponentConnection] = [ ]
        h2storage_classname = generic_hydrogen_storage.GenericHydrogenStorage.get_classname()
        connections.append(cp.ComponentConnection(L1GenericCHPRuntimeController.HydrogenSOC, h2storage_classname, generic_hydrogen_storage.GenericHydrogenStorage.HydrogenSOC))
        return connections
    def i_prepare_simulation(self) -> None:
        """ Prepares the simulation. """
        pass
    def build( self, config: L1CHPConfig ) -> None:
        self.on_time = int( config.min_operation_time / self.my_simulation_parameters.seconds_per_timestep )
        self.off_time = int( config.min_idle_time / self.my_simulation_parameters.seconds_per_timestep )
        self.SOCmin = config.min_h2_soc
        self.name = config.name
        self.source_weight = config.source_weight
        
        self.state0 = L1GenericCHPControllerState()
        self.state = L1GenericCHPControllerState()
        self.previous_state = L1GenericCHPControllerState()

    def i_save_state(self) -> None:
        self.previous_state = self.state.clone( )

    def i_restore_state(self) -> None:
        self.state = self.previous_state.clone( )

    def i_doublecheck(self, timestep: int, stsv: cp.SingleTimeStepValues) -> None:
        pass

    def i_simulate( self, timestep: int, stsv: cp.SingleTimeStepValues,  force_convergence: bool ) -> None:
        # check demand, and change state of self.has_heating_demand, and self._has_cooling_demand
        if force_convergence:
            pass
        
        l2_devicesignal = stsv.get_input_value( self.l2_DeviceSignalC )
        electricity_target = stsv.get_input_value( self.ElectricityTargetC )
        H2_SOC = stsv.get_input_value( self.HydrogenSOCC )
        
        #save reference state state0 in first iteration
        if self.state.is_first_iteration( timestep ):
            self.state0 = self.state.clone( )
        
        #return device on if minimum operation time is not fulfilled and device was on in previous state
        if ( self.state0.state == 1 and self.state0.timestep_of_last_action + self.on_time >= timestep ):
            self.state.state = 1
        #return device off if minimum idle time is not fulfilled and device was off in previous state
        elif ( self.state0.state == 0 and self.state0.timestep_of_last_action + self.off_time >= timestep ):
            self.state.state = 0
                #catch cases where hydrogen storage is close to maximum level and signals oscillate -> just turn off electrolyzer
        elif force_convergence:
            if self.state0.state == 0:
                self.state.state = 0
            else:
                self.state.deactivation( timestep )
            electricity_target = 0
        #check signal from l2 and turn on or off if it is necesary
        else:
            if ( ( l2_devicesignal == 0 ) or ( electricity_target <= 0 ) or ( H2_SOC < self.SOCmin ) ) and self.state0.state == 1:
                self.state.deactivation( timestep )
            elif ( ( l2_devicesignal == 1 ) and ( electricity_target > 0 ) and ( H2_SOC >= self.SOCmin ) ) and self.state0.state == 0:
                self.state.activation( timestep )
            
        stsv.set_output_value( self.L1DeviceSignalC, self.state.state )

    def prin1t_outpu1t(self, t_m: float, state: L1GenericCHPControllerState) -> None:
        log.information("==========================================")
        log.information(f"T m: {t_m}")
        log.information(f"State: {state}")

    def write_to_report(self) -> List[str]:
        lines: List[str] = []
        lines.append("Generic CHP L1 Controller: " + self.component_name)
        return lines



# Generic/Built-in
import random
from typing import List

# Owned
from hisim.component import Component, SingleTimeStepValues, ComponentInput, ComponentOutput
from hisim import loadtypes as lt
from hisim.simulationparameters import SimulationParameters
from typing import List
class RandomNumbers(Component):
    RandomOutput: str = "Random Numbers"

    def __init__(self, name: str, timesteps: int, minimum: float, maximum: float, my_simulation_parameters: SimulationParameters ) -> None:
        super().__init__(name, my_simulation_parameters=my_simulation_parameters)
        self.values: List[float] = []
        self.minimum = minimum
        self.maximum = maximum
        number_range = maximum - minimum
        for x in range(timesteps):
            number = minimum + random.random() * number_range
            self.values.append(number)
        self.output1 = self.add_output(self.component_name,
                                       RandomNumbers.RandomOutput,
                                       lt.LoadTypes.ANY,
                                       lt.Units.ANY)

    def i_restore_state(self) -> None:
        pass

    def i_simulate(self, timestep: int, stsv: SingleTimeStepValues,  force_convergence: bool) -> None:
        val1: float = self.values[timestep]
        stsv.set_output_value(self.output1, float(val1))

    def i_doublecheck(self, timestep: int, stsv: SingleTimeStepValues) -> None:
        pass

    def i_save_state(self) -> None:
        pass
    def i_prepare_simulation(self) -> None:
        """ Prepares the simulation. """
        pass
    def write_to_report(self) -> List[str]:
        lines =[]
        lines.append("Random number Generator: {}".format(self.component_name))
        lines.append("Minimum number: {}".format(self.minimum))
        lines.append("Maximum number: {}".format(self.maximum))
        return lines


""" Handles all the weather data processing. """

import datetime
import math
import os
from dataclasses import dataclass
from typing import List, Any
from enum import Enum
import numpy as np
import pandas as pd
import pvlib

from hisim import loadtypes as lt
from hisim import log
from hisim import utils

from hisim.component import Component, SingleTimeStepValues, ComponentOutput, ConfigBase
from hisim.simulationparameters import SimulationParameters

__authors__ = "Vitor Hugo Bellotto Zago, Noah Pflugradt"
__copyright__ = "Copyright 2021, the House Infrastructure Project"
__credits__ = ["Noah Pflugradt"]
__license__ = "MIT"
__version__ = "0.1"
__maintainer__ = "Noah Pflugradt"

""" The functions cited in this module are at some degree based on the tsib project:

[tsib-kotzur]: Kotzur, Leander, Detlef Stolten, and Hermann-Josef Wagner. Future grid load of the residential building sector.
No. RWTH-2018-231872. Lehrstuhl für Brennstoffzellen (FZ Jülich), 2019.
ID: http://hdl.handle.net/2128/21115
    http://nbn-resolving.org/resolver?verb=redirect&identifier=urn:nbn:de:0001-2019020614

The implementation of the tsib project can be found under the following repository:
https://github.com/FZJ-IEK3-VSA/tsib
"""


class WeatherDataSourceEnum(Enum):

    """ Describes where the weather data is from. Used to choose the correct reading function. """

    DWD = 1
    NSRDB = 2


class LocationEnum(Enum):

    """ contains all the locations and their corresponding directories. """

    Aachen = ("Aachen", "test-reference-years_1995-2012_1-location", "data_processed", "aachen_center", WeatherDataSourceEnum.DWD)  # noqa: invalid-name
    Bremerhaven = ("01_Bremerhaven", "test-reference-years_2015-2045_15-locations", "data_processed",  "weather_region_01", WeatherDataSourceEnum.DWD)  # noqa: invalid-name
    Rostock = ("02_Rostock", "test-reference-years_2015-2045_15-locations", "data_processed", "weather_region_02", WeatherDataSourceEnum.DWD)  # noqa: invalid-name
    Hamburg = ("03Hamburg", "test-reference-years_2015-2045_15-locations", "data_processed", "weather_region_03", WeatherDataSourceEnum.DWD)  # noqa: invalid-name
    Potsdam = ("04Potsdam", "test-reference-years_2015-2045_15-locations", "data_processed", "weather_region_04", WeatherDataSourceEnum.DWD)  # noqa: invalid-name
    Essen = ("05Essen", "test-reference-years_2015-2045_15-locations", "data_processed", "weather_region_05", WeatherDataSourceEnum.DWD)  # noqa: invalid-name
    Bad_Marienburg = ("06Bad Marienburg", "test-reference-years_2015-2045_15-locations", "data_processed", "weather_region_06", WeatherDataSourceEnum.DWD)  # noqa: invalid-name
    Kassel = ("07Kassel", "test-reference-years_2015-2045_15-locations", "data_processed", "weather_region_07", WeatherDataSourceEnum.DWD)  # noqa: invalid-name
    Braunlage = ("08Braunlage", "test-reference-years_2015-2045_15-locations", "data_processed", "weather_region_08", WeatherDataSourceEnum.DWD)  # noqa: invalid-name
    Chemnitz = ("09Chemnitz", "test-reference-years_2015-2045_15-locations", "data_processed", "weather_region_09", WeatherDataSourceEnum.DWD)  # noqa: invalid-name
    Hof = ("10Hof", "test-reference-years_2015-2045_15-locations", "data_processed", "weather_region_10", WeatherDataSourceEnum.DWD)  # noqa: invalid-name
    Fichtelberg = ("11Fichtelberg", "test-reference-years_2015-2045_15-locations", "data_processed", "weather_region_11", WeatherDataSourceEnum.DWD)  # noqa: invalid-name
    Mannheim = ("12Mannheim", "test-reference-years_2015-2045_15-locations", "data_processed", "weather_region_12", WeatherDataSourceEnum.DWD)  # noqa: invalid-name
    Muehldorf = ("13Muehldorf", "test-reference-years_2015-2045_15-locations", "data_processed", "weather_region_13", WeatherDataSourceEnum.DWD)  # noqa: invalid-name
    Stoetten = ("14Stoetten", "test-reference-years_2015-2045_15-locations", "data_processed", "weather_region_14", WeatherDataSourceEnum.DWD)  # noqa: invalid-name
    Garmisch_Partenkirchen = ("15Garmisch Partenkirchen", "test-reference-years_2015-2045_15-locations", "data_processed", "weather_region_15", WeatherDataSourceEnum.DWD)  # noqa: invalid-name
    Madrid = ("Madrid", "test-reference-years_2015-2045_15-locations", "NSRDB", "Madrid", WeatherDataSourceEnum.NSRDB)  # noqa: invalid-name
    Seville = ("Seville", "test-reference-years_2015-2045_15-locations", "NSRDB", "Seville", WeatherDataSourceEnum.NSRDB)  # noqa: invalid-name


@dataclass
class WeatherConfig(ConfigBase):

    """ Configuration class for Weather. """

    location: str
    source_path: str
    data_source: WeatherDataSourceEnum

    @classmethod
    def get_main_classname(cls):
        """ Get the name of the main class. """
        return Weather.get_full_classname()

    @classmethod
    def get_default(cls, location_entry: Any) -> Any:
        """ Gets the default configuration for Aachen. """
        path = os.path.join(utils.get_input_directory(), "weather", location_entry.value[1], location_entry.value[2], location_entry.value[3])
        config = WeatherConfig(name="Weather_1", location=location_entry.value[0], source_path=path, data_source=location_entry.value[4])
        return config


class Weather(Component):

    """ Provide thermal and solar conditions of local weather. """

    # Inputs
    # None

    # Outputs
    TemperatureOutside = "TemperatureOutside"
    DirectNormalIrradiance = "DirectNormalIrradiance"
    DiffuseHorizontalIrradiance = "DiffuseHorizontalIrradiance"
    DirectNormalIrradianceExtra = "DirectNormalIrradianceExtra"
    GlobalHorizontalIrradiance = "GlobalHorizontalIrradiance"
    Altitude = "Altitude"
    Azimuth = "Azimuth"
    ApparentZenith = "ApparentZenith"
    WindSpeed = "WindSpeed"
    Weather_Temperature_Forecast_24h = "Weather_Temperature_Forecast_24h"

    Weather_TemperatureOutside_yearly_forecast = "Weather_TemperatureOutside_yearly_forecast"
    Weather_DirectNormalIrradiance_yearly_forecast = "Weather_DirectNormalIrradiance_yearly_forecast"
    Weather_DiffuseHorizontalIrradiance_yearly_forecast = "Weather_DiffuseHorizontalIrradiance_yearly_forecast"
    Weather_DirectNormalIrradianceExtra_yearly_forecast = "Weather_DirectNormalIrradianceExtra_yearly_forecast"
    Weather_GlobalHorizontalIrradiance_yearly_forecast = "Weather_GlobalHorizontalIrradiance_yearly_forecast"
    Weather_Azimuth_yearly_forecast = "Weather_Azimuth_yearly_forecast"
    Weather_ApparentZenith_yearly_forecast = "Weather_ApparentZenith_yearly_forecast"
    Weather_WindSpeed_yearly_forecast = "Weather_WindSpeed_yearly_forecast"

    @utils.measure_execution_time
    def __init__(self, my_simulation_parameters: SimulationParameters, config: WeatherConfig):
        """ Initializes the entire class. """
        super().__init__(name="Weather", my_simulation_parameters=my_simulation_parameters)
        if my_simulation_parameters is None:
            raise Exception("Simparameters was none")
        self.last_timestep_with_update = -1
        self.weather_config = config
        self.parameter_string = my_simulation_parameters.get_unique_key()

        self.air_temperature_output: ComponentOutput = self.add_output(self.component_name, self.TemperatureOutside, lt.LoadTypes.TEMPERATURE,
                                                                       lt.Units.CELSIUS)

        self.DNI_output: ComponentOutput = self.add_output(self.component_name, self.DirectNormalIrradiance, lt.LoadTypes.IRRADIANCE,
                                                           lt.Units.WATT_PER_SQUARE_METER)

        self.DNI_extra_output: ComponentOutput = self.add_output(self.component_name, self.DirectNormalIrradianceExtra, lt.LoadTypes.IRRADIANCE,
                                                                 lt.Units.WATT_PER_SQUARE_METER)

        self.DHI_output: ComponentOutput = self.add_output(self.component_name, self.DiffuseHorizontalIrradiance, lt.LoadTypes.IRRADIANCE,
                                                           lt.Units.WATT_PER_SQUARE_METER)

        self.GHI_output: ComponentOutput = self.add_output(self.component_name, self.GlobalHorizontalIrradiance, lt.LoadTypes.IRRADIANCE,
                                                           lt.Units.WATT_PER_SQUARE_METER)

        self.altitude_output: ComponentOutput = self.add_output(self.component_name, self.Altitude, lt.LoadTypes.ANY, lt.Units.DEGREES)

        self.azimuth_output: ComponentOutput = self.add_output(self.component_name, self.Azimuth, lt.LoadTypes.ANY, lt.Units.DEGREES)

        self.apparent_zenith_output: ComponentOutput = self.add_output(self.component_name, self.ApparentZenith, lt.LoadTypes.ANY, lt.Units.DEGREES)

        self.wind_speed_output: ComponentOutput = self.add_output(self.component_name, self.WindSpeed, lt.LoadTypes.SPEED, lt.Units.METER_PER_SECOND)
        self.temperature_list: List[float]
        self.DNI_list: List[float]
        self.DNIextra_list: List[float]
        self.altitude_list: List[float]
        self.azimuth_list: List[float]
        self.wind_speed_list: List[float]
        self.GHI_list: List[float]
        self.apparent_zenith_list: List[float]
        self.DHI_list: List[float]
        self.dry_bulb_list: List[float]

    def write_to_report(self):
        """ Write configuration to the report. """
        lines = []
        lines.append("Weather")
        lines.append(self.weather_config.get_string_dict())  # type: ignore
        return lines

    def i_save_state(self) -> None:
        """ Saves the current state. """
        pass

    def i_restore_state(self) -> None:
        """ Restores the previous state. Not needed for weather. """
        pass

    def i_doublecheck(self, timestep: int, stsv: SingleTimeStepValues) -> None:
        """ Double chekc. """
        pass

    def i_simulate(self, timestep: int, stsv: SingleTimeStepValues, force_convergence: bool) -> None:
        """ Performs the simulation. """
        if self.last_timestep_with_update == timestep:
            return
        if force_convergence:
            return
        """ Performs the simulation. """
        stsv.set_output_value(self.air_temperature_output, self.temperature_list[timestep])
        stsv.set_output_value(self.DNI_output, self.DNI_list[timestep])
        stsv.set_output_value(self.DNI_extra_output, self.DNIextra_list[timestep])
        stsv.set_output_value(self.DHI_output, self.DHI_list[timestep])
        stsv.set_output_value(self.GHI_output, self.GHI_list[timestep])
        stsv.set_output_value(self.altitude_output, self.altitude_list[timestep])
        stsv.set_output_value(self.azimuth_output, self.azimuth_list[timestep])
        stsv.set_output_value(self.wind_speed_output, self.wind_speed_list[timestep])
        stsv.set_output_value(self.apparent_zenith_output, self.apparent_zenith_list[timestep])

        # set the temperature forecast
        if self.my_simulation_parameters.system_config.predictive:
            timesteps_24h = 24 * 3600 / self.my_simulation_parameters.seconds_per_timestep
            last_forecast_timestep = int(timestep + timesteps_24h)
            if last_forecast_timestep > len(self.temperature_list):
                last_forecast_timestep = len(self.temperature_list)
            # log.information( type(self.temperature))
            temperatureforecast = self.temperature_list[timestep:last_forecast_timestep]
            self.simulation_repository.set_entry(self.Weather_Temperature_Forecast_24h, temperatureforecast)
        self.last_timestep_with_update = timestep

    def i_prepare_simulation(self) -> None:
        """ Generates the lists to be used later. """
        seconds_per_timestep = self.my_simulation_parameters.seconds_per_timestep
        log.information(self.weather_config.location)
        log.information(self.weather_config.to_json())  # type: ignore
        location_dict = get_coordinates(self.weather_config.source_path)
        self.simulation_repository.set_entry("weather_location", location_dict)
        cachefound, cache_filepath = utils.get_cache_file("Weather", self.weather_config, self.my_simulation_parameters)
        if cachefound:
            # read cached files
            my_weather = pd.read_csv(cache_filepath, sep=",", decimal=".", encoding="cp1252")
            self.temperature_list = my_weather['t_out'].tolist()
            self.dry_bulb_list = self.temperature_list
            self.DHI_list = my_weather['DHI'].tolist()
            self.DNI_list = my_weather['DNI'].tolist()  # self np.float64( maybe not needed? - Noah
            self.DNIextra_list = my_weather['DNIextra'].tolist()
            self.GHI_list = my_weather['GHI'].tolist()
            self.altitude_list = my_weather['altitude'].tolist()
            self.azimuth_list = my_weather['azimuth'].tolist()
            self.apparent_zenith_list = my_weather['apparent_zenith'].tolist()
            self.wind_speed_list = my_weather['Wspd'].tolist()
        else:
            tmy_data, location = read_test_reference_year_data(weatherconfig=self.weather_config, year=self.my_simulation_parameters.year)
            DNI = self.interpolate(tmy_data['DNI'], self.my_simulation_parameters.year)
            # calculate extra terrestrial radiation- n eeded for perez array diffuse irradiance models
            dni_extra = pd.Series(pvlib.irradiance.get_extra_radiation(DNI.index), index=DNI.index)  # type: ignore
            # DNI_data = self.interpolate(tmy_data['DNI'], 2015)
            temperature = self.interpolate(tmy_data['T'], self.my_simulation_parameters.year)
            DHI = self.interpolate(tmy_data['DHI'], self.my_simulation_parameters.year)
            GHI = self.interpolate(tmy_data['GHI'], self.my_simulation_parameters.year)
            solpos = pvlib.solarposition.get_solarposition(DNI.index, location['latitude'], location['longitude'])  # type: ignore
            altitude = solpos['elevation']
            azimuth = solpos['azimuth']
            apparent_zenith = solpos['apparent_zenith']
            wind_speed = self.interpolate(tmy_data['Wspd'], self.my_simulation_parameters.year)

            if seconds_per_timestep != 60:
                self.temperature_list = temperature.resample(str(seconds_per_timestep) + "S").mean().tolist()
                self.dry_bulb_list = temperature.resample(str(seconds_per_timestep) + "S").mean().to_list()
                self.DHI_list = DHI.resample(str(seconds_per_timestep) + "S").mean().tolist()
                # np.float64( ## not sure what this is fore. python float and npfloat 64 are the same.
                self.DNI_list = DNI.resample(str(seconds_per_timestep) + "S").mean().tolist()  # )  # type: ignore
                self.DNIextra_list = dni_extra.resample(str(seconds_per_timestep) + "S").mean().tolist()
                self.GHI_list = GHI.resample(str(seconds_per_timestep) + "S").mean().tolist()
                self.altitude_list = altitude.resample(str(seconds_per_timestep) + "S").mean().tolist()
                self.azimuth_list = azimuth.resample(str(seconds_per_timestep) + "S").mean().tolist()
                self.apparent_zenith_list = apparent_zenith.resample(str(seconds_per_timestep) + "S").mean().tolist()
                self.wind_speed_list = wind_speed.resample(str(seconds_per_timestep) + "S").mean().tolist()
            else:
                self.temperature_list = temperature.tolist()
                self.dry_bulb_list = temperature.to_list()
                self.DHI_list = DHI.tolist()
                self.DNI_list = DNI.tolist()
                self.DNIextra_list = dni_extra.tolist()
                self.GHI_list = GHI.tolist()
                self.altitude_list = altitude.tolist()
                self.azimuth_list = azimuth.tolist()
                self.apparent_zenith_list = apparent_zenith.tolist()
                self.wind_speed_list = wind_speed.resample(str(seconds_per_timestep) + "S").mean().tolist()

            solardata = [self.DNI_list, self.DHI_list, self.GHI_list, self.temperature_list, self.altitude_list, self.azimuth_list,
                         self.apparent_zenith_list, self.dry_bulb_list, self.wind_speed_list, self.DNIextra_list]

            database = pd.DataFrame(np.transpose(solardata),
                                    columns=['DNI', 'DHI', 'GHI', 't_out', 'altitude', 'azimuth', 'apparent_zenith', 'DryBulb', 'Wspd', 'DNIextra'])
            database.to_csv(cache_filepath)

        # write one year forecast to simulation repository for PV processing -> if PV forecasts are needed
        if self.my_simulation_parameters.system_config.predictive:
            self.simulation_repository.set_entry(self.Weather_TemperatureOutside_yearly_forecast, self.temperature_list)
            self.simulation_repository.set_entry(self.Weather_DiffuseHorizontalIrradiance_yearly_forecast, self.DHI_list)
            self.simulation_repository.set_entry(self.Weather_DirectNormalIrradiance_yearly_forecast, self.DNI_list)
            self.simulation_repository.set_entry(self.Weather_DirectNormalIrradianceExtra_yearly_forecast, self.DNIextra_list)
            self.simulation_repository.set_entry(self.Weather_GlobalHorizontalIrradiance_yearly_forecast, self.GHI_list)
            self.simulation_repository.set_entry(self.Weather_Azimuth_yearly_forecast, self.azimuth_list)
            self.simulation_repository.set_entry(self.Weather_ApparentZenith_yearly_forecast, self.apparent_zenith_list)
            self.simulation_repository.set_entry(self.Weather_WindSpeed_yearly_forecast, self.wind_speed_list)

    def interpolate(self, pd_database: Any, year: int) -> Any:
        """ Interpolates a time series. """
        firstday = pd.Series([0.0], index=[pd.to_datetime(datetime.datetime(year - 1, 12, 31, 23, 0), utc=True).tz_convert(tz="Europe/Berlin")])
        lastday = pd.Series(pd_database[-1], index=[pd.to_datetime(datetime.datetime(year, 12, 31, 22, 59), utc=True).tz_convert(tz="Europe/Berlin")])
        pd_database = pd.concat([pd_database, firstday, lastday])
        pd_database = pd_database.sort_index()
        return pd_database.resample('1T').asfreq().interpolate(method='linear')

    def calc_sun_position(self, latitude_deg, longitude_deg, year, hoy):
        """ Calculates the Sun Position for a specific hour and location.

        :param latitude_deg: Geographical Latitude in Degrees
        :type latitude_deg: float
        :param longitude_deg: Geographical Longitude in Degrees
        :type longitude_deg: float
        :param year: year
        :type year: int
        :param hoy: Hour of the year from the start. The first hour of January is 1
        :type hoy: int
        :return: altitude, azimuth: Sun position in altitude and azimuth degrees [degrees]
        :rtype: tuple
        """
        # Convert to Radians
        latitude_rad = math.radians(latitude_deg)
        # longitude_rad = math.radians(longitude_deg)  # Note: this is never used

        # Set the date in UTC based off the hour of year and the year itself
        start_of_year = datetime.datetime(year, 1, 1, 0, 0, 0, 0)
        utc_datetime = start_of_year + datetime.timedelta(hours=hoy)

        # Angular distance of the sun north or south of the earths equator
        # Determine the day of the year.
        day_of_year = utc_datetime.timetuple().tm_yday

        # Calculate the declination angle: The variation due to the earths tilt
        # http://www.pveducation.org/pvcdrom/properties-of-sunlight/declination-angle
        declination_rad = math.radians(23.45 * math.sin((2 * math.pi / 365.0) * (day_of_year - 81)))

        # Normalise the day to 2*pi
        # There is some reason as to why it is 364 and not 365.26
        angle_of_day = (day_of_year - 81) * (2 * math.pi / 364)

        # The deviation between local standard time and true solar time
        equation_of_time = (9.87 * math.sin(2 * angle_of_day)) - (7.53 * math.cos(angle_of_day)) - (1.5 * math.sin(angle_of_day))

        # True Solar Time
        solar_time = ((utc_datetime.hour * 60) + utc_datetime.minute + (4 * longitude_deg) + equation_of_time) / 60.0

        # Angle between the local longitude and longitude where the sun is at
        # higher altitude
        hour_angle_rad = math.radians(15 * (12 - solar_time))

        # Altitude Position of the Sun in Radians
        altitude_rad = math.asin(
            math.cos(latitude_rad) * math.cos(declination_rad) * math.cos(hour_angle_rad) + math.sin(latitude_rad) * math.sin(declination_rad))

        # Azimuth Position fo the sun in radians
        azimuth_rad = math.asin(math.cos(declination_rad) * math.sin(hour_angle_rad) / math.cos(altitude_rad))

        # I don't really know what this code does, it has been imported from
        # PySolar
        if math.cos(hour_angle_rad) >= (math.tan(declination_rad) / math.tan(latitude_rad)):
            return math.degrees(altitude_rad), math.degrees(azimuth_rad)
        return math.degrees(altitude_rad), (180 - math.degrees(azimuth_rad))

    def calc_sun_position2(self, hoy: Any) -> Any:
        """ Calculates the sun position. """
        return self.altitude_list[hoy], self.azimuth_list[hoy]


def get_coordinates(filepath: str) -> Any:
    """ Reads a test reference year file and gets the GHI, DHI and DNI from it.

    Based on the tsib project @[tsib-kotzur] (Check header)
    """
    # get the correct file path
    # filepath = os.path.join(utils.HISIMPATH["weather"][location])

    # get the geoposition
    with open(filepath + ".dat", encoding="utf-8") as file_stream:
        lines = file_stream.readlines()
        location_name = lines[0].split(maxsplit=2)[2].replace('\n', '')
        lat = float(lines[1][20:37])
        lon = float(lines[2][15:30])
    return {"name": location_name, "latitude": lat, "longitude": lon}
    # self.index = pd.date_range(f"{year}-01-01 00:00:00", periods=60 * 24 * 365, freq="T", tz="Europe/Berlin")


def read_test_reference_year_data(weatherconfig: WeatherConfig, year: int) -> Any:
    """ Reads a test reference year file and gets the GHI, DHI and DNI from it.

    Based on the tsib project @[tsib-kotzur] (Check header)
    """
    # get the correct file path
    filepath = os.path.join(weatherconfig.source_path)
    if weatherconfig.data_source == WeatherDataSourceEnum.NSRDB:
        data, location_dict = read_nsrdb_data(filepath, year)
    elif weatherconfig.data_source == WeatherDataSourceEnum.DWD:
        data, location_dict = read_dwd_data(filepath, year)

    return data, location_dict


def read_dwd_data(filepath: str, year: int) -> Any:
    """ Reads the DWD data. """
    # get the geoposition
    with open(filepath + ".dat", encoding="utf-8") as file_stream:
        lines = file_stream.readlines()
        location_name = lines[0].split(maxsplit=2)[2].replace('\n', '')
        lat = float(lines[1][20:37])
        lon = float(lines[2][15:30])
    location_dict = {"name": location_name, "latitude": lat, "longitude": lon}
    # check if time series data already exists as .csv with DNI
    if os.path.isfile(filepath + ".csv"):
        data = pd.read_csv(filepath + ".csv", index_col=0, parse_dates=True, sep=";", decimal=",")
        data.index = pd.to_datetime(data.index, utc=True).tz_convert("Europe/Berlin")
    # else read from .dat and calculate DNI etc.
    else:
        # get data
        data = pd.read_csv(filepath + ".dat", sep=r"\s+", skiprows=list(range(0, 31)))
        data.index = pd.date_range(f"{year}-01-01 00:30:00", periods=8760, freq="H", tz="Europe/Berlin")
        data["GHI"] = data["D"] + data["B"]
        data = data.rename(columns={"D": "DHI", "t": "T", "WG": "Wspd", "MM": "Month", "DD": "Day", "HH": "Hour", "p": "Pressure", "WR": "Wdir"})

        # calculate direct normal
        data["DNI"] = calculate_direct_normal_radiation(data["B"], lon, lat)  # data["DNI"] = data["B"]

        # save as .csv  # data.to_csv(filepath + ".csv",sep=";",decimal=",")
    return data, location_dict


def read_nsrdb_data(filepath, year):
    """ Reads a set of NSRDB data. """
    with open(filepath + ".dat", encoding="utf-8") as file_stream:
        lines = file_stream.readlines()
        location_name = lines[0].split(maxsplit=2)[2].replace('\n', '')
        lat = float(lines[1][20:25])
        lon = float(lines[2][15:20])
    location_dict = {"name": location_name, "latitude": lat, "longitude": lon}
    # get data
    data = pd.read_csv(filepath + ".dat", sep=",", skiprows=list(range(0, 11)))
    data = data.drop(data.index[8761:8772])
    data.index = pd.date_range(f"{year}-01-01 00:30:00", periods=8760, freq="H", tz="Europe/Berlin")
    data = data.rename(
        columns={"DHI": "DHI", "Temperature": "T", "Wind Speed": "Wspd", "MM": "Month", "DD": "Day", "HH": "Hour", "Pressure": "Pressure",
                 "Wind Direction": "Wdir", "GHI": "GHI", "DNI": "DNI"})
    return data, location_dict


def calculate_direct_normal_radiation(direct_horizontal_irradation, lon, lat, zenith_tol=87.0):
    """ Calculates the direct NORMAL irradiance from the direct horizontal irradiance with the help of the PV lib.

    Based on the tsib project @[tsib-kotzur] (Check header)

    Parameters
    ----------
    direct_horizontal_irradation: pd.Series with time index
        Direct horizontal irradiance
    lon: float
        Longitude of the location
    lat: float
        Latitude of the location
    zenith_tol: float, optional
        Avoid cosines of values above a certain zenith angle of in order to avoid division by zero.

    Returns
    -------
    DNI: pd.Series

    """

    solar_pos = pvlib.solarposition.get_solarposition(direct_horizontal_irradation.index, lat, lon)
    solar_pos["apparent_zenith"][solar_pos.apparent_zenith > zenith_tol] = zenith_tol
    DNI = direct_horizontal_irradation.div(solar_pos["apparent_zenith"].apply(math.radians).apply(math.cos))
    if DNI.isnull().values.any():
        raise ValueError("Something went wrong...")
    return DNI
